import hashlib
import logging
import os
import subprocess
from importlib import import_module

import attr

from .helper import get_user
from .ssh import sshmanager
from ..resource.common import Resource, NetworkResource
from ..driver.exception import ExecutionError


class ManagedFileError(Exception):
    pass


@attr.s
class ManagedFile:
    """ The ManagedFile allows the synchronisation of a file to a remote host.
    It has to be created with the to be synced file and the target resource as
    argument:

    ::
        from labgrid.util.managedfile import ManagedFile

        ManagedFile("/tmp/examplefile", <your-resource>)

    Synchronisation is done with the sync_to_resource method.
    """
    local_path = attr.ib(
        validator=attr.validators.instance_of(str),
        converter=lambda x: os.path.realpath(str(x))
    )
    resource = attr.ib(
        validator=attr.validators.instance_of(Resource),
    )
    detect_nfs = attr.ib(default=True, validator=attr.validators.instance_of(bool))

    def __attrs_post_init__(self):
        if not os.path.isfile(self.local_path):
            raise FileNotFoundError(f"Local file {self.local_path} not found")
        self.logger = logging.getLogger(f"{self}")
        self.hash = None
        self.rpath = None
        self._on_nfs_cached = None

    def sync_to_resource(self, symlink=None):
        """sync the file to the host specified in a resource

        Raises:
            ExecutionError: if the SSH connection/copy fails
        """
        if isinstance(self.resource, NetworkResource):
            host = self.resource.host
            conn = sshmanager.open(host)

            if self._on_nfs(conn):
                self.logger.info("File %s is accessible on %s, skipping copy", self.local_path, host)
                self.rpath = os.path.dirname(self.local_path) + "/"
            else:
                self.rpath = f"{self.get_user_cache_path()}/{self.get_hash()}/"
                self.logger.info("Synchronizing %s to %s", self.local_path, host)
                conn.run_check(f"mkdir -p {self.rpath}")
                conn.put_file(
                    self.local_path,
                    f"{self.rpath}{os.path.basename(self.local_path)}"
                )
        else:
            self.rpath = os.path.dirname(self.local_path) + "/"

        if symlink is not None:
            self.logger.info("Linking")
            if isinstance(self.resource, NetworkResource):
                try:
                    conn.run_check(f"test ! -e {symlink} -o -L {symlink}")
                except ExecutionError:
                    raise ManagedFileError(f"Path {symlink} exists but is not a symlink.")
                # use short options to be compatible with busybox
                # --symbolic --force --no-dereference
                conn.run_check(f"ln -sfn {self.rpath}{os.path.basename(self.local_path)} {symlink}")
            else:
                if os.path.exists(symlink) and not os.path.islink(symlink):
                    raise ManagedFileError(f"Path {symlink} exists but is not a symlink.")
                os.symlink(f"{self.rpath}{os.path.basename(self.local_path)}", symlink)


    def _on_nfs(self, conn):
        if self._on_nfs_cached is not None:
            return self._on_nfs_cached

        if not self.detect_nfs:
            return False

        self._on_nfs_cached = False

        fmt = "inode=%i,size=%s,modified=%Y"
        # The stat command is very different on MacOs
        platform = import_module('platform')
        if platform.system() == 'Darwin':
            darwin_fmt = "inode=%i,size=%z,modified=%m"
            local = subprocess.run(["stat", "-f", darwin_fmt, self.local_path],
                                   stdout=subprocess.PIPE)
        else:
            local = subprocess.run(["stat", "--format", fmt, self.local_path],
                                   stdout=subprocess.PIPE)

        if local.returncode != 0:
            self.logger.debug("local: stat: unsuccessful error code %d", local.returncode)
            return False

        remote = conn.run(f"stat --format '{fmt}' {self.local_path}",
                          decodeerrors="backslashreplace")
        if remote[2] != 0:
            self.logger.debug("remote: stat: unsuccessful error code %d", remote[2])
            return False

        localout = local.stdout.decode("utf-8", "backslashreplace").split('\n')
        localout.pop() # remove trailing empty element

        if remote[0] != localout:
            self.logger.debug("state: local (%s) and remote (%s) output don't match",
                              remote[0], localout)
            return False

        self.rpath = os.path.dirname(self.local_path) + "/"
        self._on_nfs_cached = True

        return True

    def get_remote_path(self):
        """Retrieve the remote file path

        Returns:
            str: path to the file on the remote host
        """
        if isinstance(self.resource, NetworkResource):
            if self.rpath is None:
                raise ManagedFileError("sync_to_resource() needs to be called before the remote-path can be retrieved")

            return f"{self.rpath}{os.path.basename(self.local_path)}"

        return self.local_path

    def get_hash(self):
        """Retrieve the hash of the file

        Returns:
            str: SHA256 hexdigest of the file
        """

        if self.hash is not None:
            return self.hash

        hasher = hashlib.sha256()
        with open(self.local_path, 'rb') as f:
            for block in iter(lambda: f.read(1048576), b''):
                hasher.update(block)
        self.hash = hasher.hexdigest()

        return self.hash

    def get_user_cache_path(self):
        return f"/var/cache/labgrid/{get_user()}"



from unittest.mock import MagicMock, patch
import sys

import pytest

# mock laam before importing drivers
laam_mock = MagicMock()
laam_mock.exceptions.LAAError = type('LAAError', (Exception,), {})
sys.modules['laam'] = laam_mock
sys.modules['laam.exceptions'] = laam_mock.exceptions

from labgrid.resource.laa import (
    LAASerialPort, LAAPowerPort, LAAUSBGadgetMassStorage, LAAUSBPort,
    LAAButtonPort, LAALed, LAATempSensor, LAAWattMeter, LAAProvider,
)
from labgrid.driver.laadriver import (
    _get_laam,
    LAASerialDriver, LAAPowerDriver, LAAUSBGadgetMassStorageDriver,
    LAAUSBDriver, LAAButtonDriver, LAALedDriver, LAATempDriver,
    LAAWattDriver, LAAProviderDriver,
)
from labgrid.driver.exception import ExecutionError


class TestLAAResources:
    def test_serial_port(self, target):
        r = LAASerialPort(target, 'serial', laa_identity='test-laa', serial_name='ttymxc3')
        assert r.laa_identity == 'test-laa'
        assert r.serial_name == 'ttymxc3'

    def test_power_port(self, target):
        r = LAAPowerPort(
            target, 'power',
            laa_identity='test-laa',
            power_on=[('5v', 'on')],
            power_off=[('5v', 'off')],
        )
        assert r.power_on == [('5v', 'on')]
        assert r.power_cycle is None

    def test_usb_port(self, target):
        r = LAAUSBPort(target, 'usb', laa_identity='test-laa', usb_ports=[1, 2])
        assert r.usb_ports == [1, 2]

    def test_button_port(self, target):
        r = LAAButtonPort(target, 'button', laa_identity='test-laa', buttons=['power', 'reset'])
        assert r.buttons == ['power', 'reset']

    def test_led(self, target):
        r = LAALed(target, 'led', laa_identity='test-laa')
        assert r.laa_identity == 'test-laa'

    def test_temp_sensor(self, target):
        r = LAATempSensor(target, 'temp', laa_identity='test-laa')
        assert r.laa_identity == 'test-laa'

    def test_watt_meter(self, target):
        r = LAAWattMeter(target, 'watt', laa_identity='test-laa')
        assert r.laa_identity == 'test-laa'

    def test_power_port_bad_sequence(self, target):
        with pytest.raises(ValueError, match="must be .vbus, state."):
            LAAPowerPort(
                target, 'power',
                laa_identity='test-laa',
                power_on=[('5v',)],
                power_off=[('5v', 'off')],
            )

    def test_power_port_bad_entry_type(self, target):
        with pytest.raises(ValueError, match="must be .str, str."):
            LAAPowerPort(
                target, 'power',
                laa_identity='test-laa',
                power_on=[(5, 'on')],
                power_off=[('5v', 'off')],
            )


class TestGetLaam:
    def test_missing_laam(self):
        with patch('labgrid.driver.laadriver.import_module',
                   side_effect=ModuleNotFoundError("No module named 'laam'")):
            with pytest.raises(ModuleNotFoundError, match="install labgrid"):
                _get_laam()


class TestLAASerialDriver:
    def test_create(self, target):
        LAASerialPort(target, 'serial', laa_identity='test-laa', serial_name='ttymxc3')
        d = LAASerialDriver(target, 'driver')
        assert isinstance(d, LAASerialDriver)

    def test_activate_deactivate(self, target):
        LAASerialPort(target, 'serial', laa_identity='test-laa', serial_name='ttymxc3')
        d = LAASerialDriver(target, 'driver')
        target.activate(d)

        laam_mock.LAA.assert_called_with('test-laa')
        laam_mock.LAA().serials.connect_pexpect.assert_called_with('ttymxc3')

        target.deactivate(d)
        assert d._laa is None
        assert d._conn is None

    def test_read_encodes_str_to_bytes(self, target):
        LAASerialPort(target, 'serial', laa_identity='test-laa', serial_name='ttymxc3')
        d = LAASerialDriver(target, 'driver')
        target.activate(d)

        # ConnectPexpect returns str, driver must encode to bytes
        d._conn.read_nonblocking.return_value = "hello"
        result = d._read(size=5, timeout=1.0)
        assert result == b"hello"
        assert isinstance(result, bytes)

    def test_write(self, target):
        LAASerialPort(target, 'serial', laa_identity='test-laa', serial_name='ttymxc3')
        d = LAASerialDriver(target, 'driver')
        target.activate(d)

        d._write(b"test")
        d._conn.send.assert_called_with(b"test")


class TestLAAPowerDriver:
    def test_create(self, target):
        LAAPowerPort(
            target, 'power',
            laa_identity='test-laa',
            power_on=[('5v', 'on')],
            power_off=[('5v', 'off')],
        )
        d = LAAPowerDriver(target, 'driver')
        assert isinstance(d, LAAPowerDriver)

    def test_deactivate(self, target):
        LAAPowerPort(
            target, 'power',
            laa_identity='test-laa',
            power_on=[('5v', 'on')],
            power_off=[('5v', 'off')],
        )
        d = LAAPowerDriver(target, 'driver')
        target.activate(d)
        target.deactivate(d)
        assert d._laa is None

    def test_on_off(self, target):
        LAAPowerPort(
            target, 'power',
            laa_identity='test-laa',
            power_on=[('5v', 'on'), ('3v3', 'on')],
            power_off=[('3v3', 'off'), ('5v', 'off')],
        )
        d = LAAPowerDriver(target, 'driver')
        target.activate(d)

        d.on()
        calls = d._laa.laacli.power.call_args_list
        assert calls[-2][0] == ('5v', 'on')
        assert calls[-1][0] == ('3v3', 'on')

        d._laa.laacli.power.reset_mock()
        d.off()
        calls = d._laa.laacli.power.call_args_list
        assert calls[0][0] == ('3v3', 'off')
        assert calls[1][0] == ('5v', 'off')

    def test_cycle_default(self, target, mocker):
        mocker.patch('time.sleep')
        LAAPowerPort(
            target, 'power',
            laa_identity='test-laa',
            power_on=[('5v', 'on')],
            power_off=[('5v', 'off')],
        )
        d = LAAPowerDriver(target, 'driver')
        target.activate(d)
        d.cycle()

    def test_cycle_explicit(self, target):
        LAAPowerPort(
            target, 'power',
            laa_identity='test-laa',
            power_on=[('5v', 'on')],
            power_off=[('5v', 'off')],
            power_cycle=[('5v', 'reset')],
        )
        d = LAAPowerDriver(target, 'driver')
        target.activate(d)
        d.cycle()
        d._laa.laacli.power.assert_called_with('5v', 'reset')

    def test_power_error(self, target):
        LAAPowerPort(
            target, 'power',
            laa_identity='test-laa',
            power_on=[('5v', 'on')],
            power_off=[('5v', 'off')],
        )
        d = LAAPowerDriver(target, 'driver')
        target.activate(d)

        d._laa.laacli.power.side_effect = laam_mock.exceptions.LAAError("fail")
        with pytest.raises(ExecutionError):
            d.on()


class TestLAAUSBGadgetMassStorageDriver:
    def test_create(self, target):
        LAAUSBGadgetMassStorage(
            target, 'usbg_ms', laa_identity='test-laa', image='boot.img',
        )
        d = LAAUSBGadgetMassStorageDriver(target, 'driver')
        assert isinstance(d, LAAUSBGadgetMassStorageDriver)

    def test_deactivate(self, target):
        LAAUSBGadgetMassStorage(
            target, 'usbg_ms', laa_identity='test-laa', image='boot.img',
        )
        d = LAAUSBGadgetMassStorageDriver(target, 'driver')
        target.activate(d)
        target.deactivate(d)
        assert d._laa is None

    def test_on(self, target):
        LAAUSBGadgetMassStorage(
            target, 'usbg_ms', laa_identity='test-laa', image='boot.img',
        )
        d = LAAUSBGadgetMassStorageDriver(target, 'driver')
        target.activate(d)

        d.on()
        d._laa.laacli.usbg_ms.assert_called_with('on', 'boot.img')

    def test_off(self, target):
        LAAUSBGadgetMassStorage(
            target, 'usbg_ms', laa_identity='test-laa', image='boot.img',
        )
        d = LAAUSBGadgetMassStorageDriver(target, 'driver')
        target.activate(d)

        d.off()
        d._laa.laacli.usbg_ms.assert_called_with('off', '')

    def test_on_error(self, target):
        LAAUSBGadgetMassStorage(
            target, 'usbg_ms', laa_identity='test-laa', image='boot.img',
        )
        d = LAAUSBGadgetMassStorageDriver(target, 'driver')
        target.activate(d)

        d._laa.laacli.usbg_ms.side_effect = laam_mock.exceptions.LAAError("fail")
        with pytest.raises(ExecutionError):
            d.on()

    def test_off_error(self, target):
        LAAUSBGadgetMassStorage(
            target, 'usbg_ms', laa_identity='test-laa', image='boot.img',
        )
        d = LAAUSBGadgetMassStorageDriver(target, 'driver')
        target.activate(d)

        d._laa.laacli.usbg_ms.side_effect = laam_mock.exceptions.LAAError("fail")
        with pytest.raises(ExecutionError):
            d.off()


class TestLAAUSBDriver:
    def test_deactivate(self, target):
        LAAUSBPort(target, 'usb', laa_identity='test-laa', usb_ports=[1])
        d = LAAUSBDriver(target, 'driver')
        target.activate(d)
        target.deactivate(d)
        assert d._laa is None

    def test_on_off(self, target):
        LAAUSBPort(target, 'usb', laa_identity='test-laa', usb_ports=[1, 2])
        d = LAAUSBDriver(target, 'driver')
        target.activate(d)

        d.on()
        calls = d._laa.laacli.usb.call_args_list
        assert calls[-2][0] == (1, 'on')
        assert calls[-1][0] == (2, 'on')

        d._laa.laacli.usb.reset_mock()
        d.off()
        calls = d._laa.laacli.usb.call_args_list
        assert calls[0][0] == (1, 'off')
        assert calls[1][0] == (2, 'off')

    def test_off_error(self, target):
        LAAUSBPort(target, 'usb', laa_identity='test-laa', usb_ports=[1])
        d = LAAUSBDriver(target, 'driver')
        target.activate(d)

        d._laa.laacli.usb.side_effect = laam_mock.exceptions.LAAError("fail")
        with pytest.raises(ExecutionError):
            d.off()

    def test_on_error(self, target):
        LAAUSBPort(target, 'usb', laa_identity='test-laa', usb_ports=[1])
        d = LAAUSBDriver(target, 'driver')
        target.activate(d)

        d._laa.laacli.usb.side_effect = laam_mock.exceptions.LAAError("fail")
        with pytest.raises(ExecutionError):
            d.on()


class TestLAAButtonDriver:
    def test_deactivate(self, target):
        LAAButtonPort(target, 'button', laa_identity='test-laa', buttons=['power'])
        d = LAAButtonDriver(target, 'driver')
        target.activate(d)
        target.deactivate(d)
        assert d._laa is None

    def test_press_release(self, target):
        LAAButtonPort(target, 'button', laa_identity='test-laa', buttons=['power', 'reset'])
        d = LAAButtonDriver(target, 'driver')
        target.activate(d)

        d.press('power')
        d._laa.laacli.button.assert_called_with('power', 'on')

        d.release('power')
        d._laa.laacli.button.assert_called_with('power', 'off')

    def test_invalid_button(self, target):
        LAAButtonPort(target, 'button', laa_identity='test-laa', buttons=['power'])
        d = LAAButtonDriver(target, 'driver')
        target.activate(d)

        with pytest.raises(ExecutionError):
            d.press('nonexistent')

    def test_invalid_button_release(self, target):
        LAAButtonPort(target, 'button', laa_identity='test-laa', buttons=['power'])
        d = LAAButtonDriver(target, 'driver')
        target.activate(d)

        with pytest.raises(ExecutionError):
            d.release('nonexistent')

    def test_press_error(self, target):
        LAAButtonPort(target, 'button', laa_identity='test-laa', buttons=['power'])
        d = LAAButtonDriver(target, 'driver')
        target.activate(d)

        d._laa.laacli.button.side_effect = laam_mock.exceptions.LAAError("fail")
        with pytest.raises(ExecutionError):
            d.press('power')

    def test_release_error(self, target):
        LAAButtonPort(target, 'button', laa_identity='test-laa', buttons=['power'])
        d = LAAButtonDriver(target, 'driver')
        target.activate(d)

        d._laa.laacli.button.side_effect = laam_mock.exceptions.LAAError("fail")
        with pytest.raises(ExecutionError):
            d.release('power')


class TestLAALedDriver:
    def test_deactivate(self, target):
        LAALed(target, 'led', laa_identity='test-laa')
        d = LAALedDriver(target, 'driver')
        target.activate(d)
        target.deactivate(d)
        assert d._laa is None

    def test_on_off(self, target):
        LAALed(target, 'led', laa_identity='test-laa')
        d = LAALedDriver(target, 'driver')
        target.activate(d)

        d.on()
        d._laa.laacli.led.assert_called_with('on')

        d.off()
        d._laa.laacli.led.assert_called_with('off')

    def test_error(self, target):
        LAALed(target, 'led', laa_identity='test-laa')
        d = LAALedDriver(target, 'driver')
        target.activate(d)

        d._laa.laacli.led.side_effect = laam_mock.exceptions.LAAError("fail")
        with pytest.raises(ExecutionError):
            d.on()

    def test_off_error(self, target):
        LAALed(target, 'led', laa_identity='test-laa')
        d = LAALedDriver(target, 'driver')
        target.activate(d)

        d._laa.laacli.led.side_effect = laam_mock.exceptions.LAAError("fail")
        with pytest.raises(ExecutionError):
            d.off()


class TestLAATempDriver:
    def test_deactivate(self, target):
        LAATempSensor(target, 'temp', laa_identity='test-laa')
        d = LAATempDriver(target, 'driver')
        target.activate(d)
        target.deactivate(d)
        assert d._laa is None

    def test_get_temp(self, target):
        LAATempSensor(target, 'temp', laa_identity='test-laa')
        d = LAATempDriver(target, 'driver')
        target.activate(d)

        d._laa.laacli.temp.return_value = {'temp': 42.5}
        result = d.get_temp('dut')
        d._laa.laacli.temp.assert_called_with('dut')
        assert result == {'temp': 42.5}

    def test_error(self, target):
        LAATempSensor(target, 'temp', laa_identity='test-laa')
        d = LAATempDriver(target, 'driver')
        target.activate(d)

        d._laa.laacli.temp.side_effect = laam_mock.exceptions.LAAError("fail")
        with pytest.raises(ExecutionError):
            d.get_temp('dut')


class TestLAAWattDriver:
    def test_deactivate(self, target):
        LAAWattMeter(target, 'watt', laa_identity='test-laa')
        d = LAAWattDriver(target, 'driver')
        target.activate(d)
        target.deactivate(d)
        assert d._laa is None

    def test_get_watts(self, target):
        LAAWattMeter(target, 'watt', laa_identity='test-laa')
        d = LAAWattDriver(target, 'driver')
        target.activate(d)

        d._laa.laacli.watt.return_value = {'power': 2.5}
        result = d.get_watts('5v')
        d._laa.laacli.watt.assert_called_with('5v')
        assert result == {'power': 2.5}

    def test_error(self, target):
        LAAWattMeter(target, 'watt', laa_identity='test-laa')
        d = LAAWattDriver(target, 'driver')
        target.activate(d)

        d._laa.laacli.watt.side_effect = laam_mock.exceptions.LAAError("fail")
        with pytest.raises(ExecutionError):
            d.get_watts('5v')


class TestLAAProviderDriver:
    def test_deactivate(self, target):
        LAAProvider(target, 'provider', laa_identity='test-laa')
        d = LAAProviderDriver(target, 'driver')
        target.activate(d)
        target.deactivate(d)
        assert d._laa is None

    def test_create(self, target):
        LAAProvider(target, 'provider', laa_identity='test-laa')
        d = LAAProviderDriver(target, 'driver')
        assert isinstance(d, LAAProviderDriver)

    def test_stage_local(self, target, tmp_path):
        LAAProvider(target, 'provider', laa_identity='test-laa')
        d = LAAProviderDriver(target, 'driver')
        target.activate(d)

        test_file = tmp_path / "kernel.img"
        test_file.write_text("data")

        result = d.stage(str(test_file))
        d._laa.files.push.assert_called_with('kernel.img', str(test_file))
        assert result == 'kernel.img'

    def test_stage_url(self, target, mocker):
        import io

        LAAProvider(target, 'provider', laa_identity='test-laa')
        d = LAAProviderDriver(target, 'driver')
        target.activate(d)

        mock_response = MagicMock(wraps=io.BytesIO(b"data"))
        mock_response.__enter__ = MagicMock(return_value=mock_response)
        mock_response.__exit__ = MagicMock(return_value=False)
        mocker.patch('labgrid.driver.laadriver.urlopen', return_value=mock_response)

        d._laa.files.push.reset_mock()
        result = d.stage('https://example.com/images/kernel.img')
        assert result == 'kernel.img'
        d._laa.files.push.assert_called_once()
        name, path = d._laa.files.push.call_args[0]
        assert name == 'kernel.img'

    def test_stage_missing_file(self, target):
        LAAProvider(target, 'provider', laa_identity='test-laa')
        d = LAAProviderDriver(target, 'driver')
        target.activate(d)

        with pytest.raises(ExecutionError, match="not found"):
            d.stage('/nonexistent/file.img')

    def test_list(self, target):
        LAAProvider(target, 'provider', laa_identity='test-laa')
        d = LAAProviderDriver(target, 'driver')
        target.activate(d)

        d._laa.files.list.return_value = ['kernel.img', 'dtb.dtb']
        result = d.list()
        d._laa.files.list.assert_called_once()
        assert result == ['kernel.img', 'dtb.dtb']

    def test_remove(self, target):
        LAAProvider(target, 'provider', laa_identity='test-laa')
        d = LAAProviderDriver(target, 'driver')
        target.activate(d)

        d.remove('kernel.img')
        d._laa.files.remove.assert_called_with('kernel.img')

    def test_remove_error(self, target):
        LAAProvider(target, 'provider', laa_identity='test-laa')
        d = LAAProviderDriver(target, 'driver')
        target.activate(d)

        d._laa.files.remove.side_effect = laam_mock.exceptions.LAAError("fail")
        with pytest.raises(ExecutionError):
            d.remove('kernel.img')



from .atomic import atomic_replace
from .dict import diff_dict, flat_dict, filter_dict, find_dict
from .expect import PtxExpect
from .timeout import Timeout
from .marker import gen_marker
from .yaml import load, dump
from .ssh import sshmanager
from .helper import get_free_port, get_user, re_vt100
from .version import labgrid_version



import time

from labgrid.driver import ExternalConsoleDriver


class TestExternalConsoleDriver:
    def test_create(self, target):
        d = ExternalConsoleDriver(target, "console", cmd="cat")
        assert isinstance(d, ExternalConsoleDriver)

    def test_communicate(self, target):
        data = b"test\ndata"
        d = ExternalConsoleDriver(target, "console", cmd="cat")
        target.activate(d)
        d.write(data)
        time.sleep(0.1)
        assert d.read(1024) == data  # assert written data is read
        d.write(data)
        time.sleep(0.1)
        assert d.read(5, max_size=5) == data[:5]  # assert max_size limits read bytes
        d.close()

    def test_chunking(self, target):
        data = b"even sized\ndata."
        d = ExternalConsoleDriver(target, 'console', cmd='cat')
        target.activate(d)
        d.txchunk = 2
        d.txdelay = 0.01
        d.write(data)
        time.sleep(0.1)
        assert d.read(len(data)) == data  # assert written data is read

        data = b"odd sized\ndata."
        d.write(data)
        time.sleep(0.1)
        assert d.read(len(data)) == data  # assert written data is read

        data = b"odd chunk\nsize."
        d.txchunk = 3
        d.write(data)
        time.sleep(0.1)
        assert d.read(len(data)) == data  # assert written data is read
        d.close()



import enum

import attr

from labgrid import target_factory, step
from labgrid.driver import BareboxDriver, ShellDriver
from labgrid.protocol import PowerProtocol
from labgrid.strategy import Strategy, never_retry


@attr.s(eq=False)
class StrategyError(Exception):
    msg = attr.ib(validator=attr.validators.instance_of(str))


class Status(enum.Enum):
    unknown = 0
    barebox = 1
    shell = 2


@target_factory.reg_driver
@attr.s(eq=False)
class BareboxRebootStrategy(Strategy):
    """A Strategy to switch to barebox or shell and back via reboot

    In your env.yaml, simply instantiate the strategy without parameters, for example:

    .. code-block:: yaml
        targets:
            main:
                resources:
                  RawSerialPort:
                    port: "/dev/ttyUSB0"
                drivers:
                  ManualPowerDriver:
                    name: "example"
                  SerialDriver {}
                  BareboxDriver:
                    prompt: 'barebox@[^:]+:[^ ]+ '
                    bootstring: '\[[ .0-9]+\] '
                  ShellDriver:
                    prompt: 'root@\w+:[^ ]+ '
                    login_prompt: ' login: '
                    username: 'root'
                  BareboxRebootStrategy: {}
    """

    # pull in the according drivers into self.*
    bindings = {
        "power": PowerProtocol,
        "barebox": BareboxDriver,
        "shell": ShellDriver,
    }

    status = attr.ib(default=Status.unknown)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    @never_retry
    @step(args=["new_status"])
    def transition(self, new_status, *, step):
        if not isinstance(new_status, Status):
            new_status = Status[new_status]

        if new_status == Status.unknown:
            raise StrategyError(f"can not transition to {new_status}")

        elif self.status == new_status:
            step.skip("nothing to do")
            return

        elif self.status == Status.unknown:
            # unknown state after initialization, cycle power
            self.target.activate(self.power)
            self.power.cycle()

            # we can then let the according driver do the rest to get us into a barebox or shell:
            if new_status == Status.barebox:
                self.target.activate(self.barebox)
            elif new_status == Status.shell:
                # (this assumes that barebox will autoboot to the shell after a timeout)
                self.target.activate(self.shell)

        elif self.status == Status.barebox and new_status == Status.shell:
            # in barebox: boot the default target and hope it will get us to a shell :->
            self.barebox.boot("")
            self.barebox.await_boot()
            self.target.activate(self.shell)

        elif self.status == Status.shell and new_status == Status.barebox:
            # in shell: we can simply reboot to get a barebox
            # make sure run() reads the shell prompt and returns
            self.shell.run("(sleep 1;reboot)&")
            self.target.activate(self.barebox)

        else:
            raise StrategyError(f"no transition found from {self.status} to {new_status}")

        self.status = new_status



"""
Driver for Digital Loggers PDU that use the legacy HTTP API.
Tested with Web Power Switch 7.
"""

import re
import requests


def power_set(host, port, index, value):
    assert port is None

    index = int(index)
    value = "ON" if value else "OFF"
    host = f"{host}/outlet?{index}={value}"
    r = requests.get(host)
    r.raise_for_status()


def power_get(host, port, index):
    assert port is None

    index = int(index)
    host = f"{host}/status"
    r = requests.get(host)
    r.raise_for_status()

    # Basically, an HTML page is returned, whose body contents are like:
    # <div id="state">ff</div><div id="lock">00</div><div id="perm">ff</div>
    # We're interested in the value of state div - 'ff' above. And the
    # status is basically the corresponding outlet bit index. So, outlet
    # 1 is the 0th bit, with 1 for ON and 0 for OFF.
    statuses = re.findall(r'state">(\w+)<', r.text)[0]
    return bool(int(statuses, 16) & (1 << (index - 1)))



def get_channel_info(driver, channel):
    info = {}
    info["SELECTED"] = driver.get_str(f"SELect:CH{channel:d)}?").strip()
    info["BWLIMIT"] = driver.get_str(f":CH{channel:d}:BANdwidth?").strip()
    info["COUPLING"] = driver.get_str(f":CH{channel:d}:COUPling?").strip()
    info["INVERT"] = driver.get_str(f":CH{channel:d}:INVert?").strip()
    info["POSITION"] = driver.get_decimal(f":CH{channel:d}:POSition?")
    info["PROBE"] = driver.get_decimal(f":CH{channel:d}:PRObe?")
    info["SCALE"] = driver.get_decimal(f":CH{channel:d}:SCAle?")
    return info


def get_channel_values(driver, channel):
    driver.command(f":MEASUrement:IMMed:SOUrce1 CH{channel:d}")
    info = {}
    driver.command(":MEASUrement:IMMed:TYPe MEAN")
    info["MEAN"] = driver.get_decimal(":MEASUrement:IMMed:VALue?")
    driver.command(":MEASUrement:IMMed:TYPe FREQuency")
    info["FREQUENCY"] = driver.get_decimal(":MEASUrement:IMMed:VALue?")
    driver.command(":MEASUrement:IMMed:TYPe PK2pk")
    info["PK2PK"] = driver.get_decimal(":MEASUrement:IMMed:VALue?")
    driver.command(":MEASUrement:IMMed:TYPe MINImum")
    info["MINIMUM"] = driver.get_decimal(":MEASUrement:IMMed:VALue?")
    driver.command(":MEASUrement:IMMed:TYPe MAXImum")
    info["MAXIMUM"] = driver.get_decimal(":MEASUrement:IMMed:VALue?")
    return info


def get_screenshot_tiff(driver):
    driver.command(":hardcopy:format tiff")
    tiff = driver.query(":hardcopy start", raw=True)
    return tiff



import pytest
import requests

from labgrid.driver import HttpDigitalOutputDriver
from labgrid.resource import HttpDigitalOutput


@pytest.fixture(scope="function")
def mock_server(mocker):
    state = '"Unknown"'

    def request(method, url, data=None):
        nonlocal state
        state = data
        return mocker.MagicMock()

    def get(url):
        r = mocker.MagicMock()
        r.text = state
        return r

    mock_request = mocker.patch("requests.request")
    mock_request.side_effect = request
    mock_get = mocker.patch("requests.get")
    mock_get.side_effect = get

    return (mock_request, mock_get)


def _make_http_driver(target, with_tls, with_regex, separate_get, match_error):
    scheme = "https" if with_tls else "http"
    url = f"{scheme}://host.example/set"
    url_get = f"{scheme}://host.example/get" if separate_get else ""

    body_get_asserted = ".*n.*" if with_regex else ""
    body_get_deasserted = ".*ff.*" if with_regex else ""

    if match_error:
        body_get_asserted = "--- DOES NOT MATCH ---"
        body_get_deasserted = "--- DOES NOT MATCH EITHER ---"

    dig_out_res = HttpDigitalOutput(
        target,
        name=None,
        url=url,
        body_asserted='"On"',
        body_deasserted='"Off"',
        method="PUT",
        url_get=url_get,
        body_get_asserted=body_get_asserted,
        body_get_deasserted=body_get_deasserted,
    )

    http_driver = HttpDigitalOutputDriver(target, name=None)
    target.activate(http_driver)

    return http_driver


@pytest.mark.parametrize(
    "asserted,with_tls,with_regex,separate_get",
    [
        (False, False, False, False),
        (True, False, False, False),
        (True, True, False, False),
        (True, False, True, False),
        (False, False, True, False),
        (True, False, False, True),
        (True, True, True, True),
    ],
)
def test_set_get(asserted, with_tls, with_regex, separate_get, target, mock_server):
    http_driver = _make_http_driver(target, with_tls, with_regex, separate_get, False)
    mock_request, mock_get = mock_server

    data = '"On"' if asserted else '"Off"'
    scheme = "https" if with_tls else "http"
    port = 443 if with_tls else 80
    get_endpoint = "get" if separate_get else "set"

    set_url = f"{scheme}://host.example:{port}/set"
    get_url = f"{scheme}://host.example:{port}/{get_endpoint}"

    http_driver.set(asserted)
    mock_request.assert_called_once_with("PUT", set_url, data=data)

    assert http_driver.get() == asserted
    mock_get.assert_called_once_with(get_url)


def test_match_exception(target, mock_server):
    http_driver = _make_http_driver(target, False, False, False, True)
    mock_request, mock_get = mock_server

    http_driver.set(True)
    with pytest.raises(Exception):
        http_driver.get()



import pytest


@pytest.fixture(scope="function")
def in_bootloader(strategy, capsys):
    with capsys.disabled():
        strategy.transition("uboot")


@pytest.fixture(scope="function")
def in_shell(strategy, capsys):
    with capsys.disabled():
        strategy.transition("shell")


def test_uboot(target, in_bootloader):
    # command = target.get_driver(CommandProtocol)
    command = target.get_driver("UBootDriver")

    stdout, stderr, returncode = command.run("version")
    assert returncode == 0
    assert stdout
    assert not stderr
    assert "U-Boot" in "\n".join(stdout)


def test_shell(target, in_shell):
    # command = target.get_driver(CommandProtocol)
    command = target.get_driver("ShellDriver")
    stdout, stderr, returncode = command.run("cat /proc/version")
    assert returncode == 0
    assert stdout
    assert not stderr
    assert "Linux" in stdout[0]


def test_uboot_2(target, in_bootloader):
    command = target.get_driver("UBootDriver")

    stdout, stderr, returncode = command.run("version")
    assert returncode == 0
    assert stdout
    assert not stderr
    assert "U-Boot" in "\n".join(stdout)



import subprocess

import attr

from ..exceptions import InvalidConfigError
from ..factory import target_factory
from ..protocol import VideoProtocol
from .common import Driver


@target_factory.reg_driver
@attr.s(eq=False)
class USBVideoDriver(Driver, VideoProtocol):
    bindings = {
        "video": {"USBVideo", "NetworkUSBVideo"},
    }

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self._prepared = False

    def get_qualities(self):
        match = (self.video.vendor_id, self.video.model_id)
        if match == (0x046d, 0x082d):
            return ("mid", [
                ("low", "video/x-h264,width=640,height=360,framerate=5/1"),
                ("mid", "video/x-h264,width=1280,height=720,framerate=15/2"),
                ("high", "video/x-h264,width=1920,height=1080,framerate=10/1"),
                ])
        elif match == (0x046d, 0x0892):
            return ("mid", [
                ("low", "image/jpeg,width=640,height=360,framerate=5/1"),
                ("mid", "image/jpeg,width=1280,height=720,framerate=15/2"),
                ("high", "image/jpeg,width=1920,height=1080,framerate=10/1"),
                ])
        elif match == (0x046d, 0x08e5): # Logitech HD Pro Webcam C920
            return ("mid", [
                ("low", "image/jpeg,width=640,height=360,framerate=5/1"),
                ("mid", "image/jpeg,width=1280,height=720,framerate=15/2"),
                ("high", "image/jpeg,width=1920,height=1080,framerate=10/1"),
                ])
        elif match == (0x1224, 0x2825): # LogiLink UA0371
            return ("mid", [
                ("low", "image/jpeg,width=640,height=480,framerate=30/1"),
                ("mid", "image/jpeg,width=1280,height=720,framerate=30/1"),
                ("high", "image/jpeg,width=1920,height=1080,framerate=30/1"),
                ])
        elif match == (0x05a3, 0x9331): # WansView Webcam 102
            return ("mid", [
                ("low","video/x-h264,width=640,height=360,framerate=30/1"),
                ("mid","video/x-h264,width=1280,height=720,framerate=30/1"),
                ("high","video/x-h264,width=1920,height=1080,framerate=30/1"),
                ])
        elif match == (0x534d, 0x2109): # MacroSilicon
            return ("mid", [
                ("low", "image/jpeg,width=720,height=480,framerate=10/1"),
                ("mid", "image/jpeg,width=1280,height=720,framerate=10/1"),
                ("high", "image/jpeg,width=1920,height=1080,framerate=10/1"),
                ])
        elif match == (0x1d6c, 0x0103): # HD 2MP WEBCAM
            return ("mid", [
                ("low", "video/x-h264,width=640,height=480,framerate=25/1"),
                ("mid", "video/x-h264,width=1280,height=720,framerate=25/1"),
                ("high", "video/x-h264,width=1920,height=1080,framerate=25/1"),
                ])
        elif match == (0x0c45, 0x636b): # LogiLink UA0379 / Microdia
            return ("mid", [
                ("low", "image/jpeg,width=640,height=480,pixel-aspect-ratio=1/1,framerate=30/1"),
                ("mid", "image/jpeg,width=1280,height=720,pixel-aspect-ratio=1/1,framerate=30/1"),
                ("high", "image/jpeg,width=1920,height=1080,pixel-aspect-ratio=1/1,framerate=30/1"),
                ])
        elif match == (0x0c45, 0x636d): # AUKEY PC-LM1E
            return ("mid", [
                ("low", "image/jpeg,width=640,height=480,pixel-aspect-ratio=1/1,framerate=30/1"),
                ("mid", "image/jpeg,width=864,height=480,pixel-aspect-ratio=1/1,framerate=30/1"),
                ("high", "image/jpeg,width=1280,height=1024,pixel-aspect-ratio=1/1,framerate=30/1"),
                ])
        elif match == (0x345f, 0x2130):  # UltraSemi USB3 Video
            return ("mid", [
                ("low", "image/jpeg,width=1024,height=768,pixel-aspect-ratio=1/1,framerate=30/1"),
                ("mid", "image/jpeg,width=1920,height=1080,pixel-aspect-ratio=1/1,framerate=30/1"),
                ("high", "image/jpeg,width=3840,height=2160,pixel-aspect-ratio=1/1,framerate=18/1"),
                ])
        elif match == (0x0c45, 0x6366): # Anivia W8 - Sonix Technology Co. Ltd.
            return ("mid", [
                ("low", "video/x-raw,format=YUY2,width=640,height=400,pixel-aspect-ratio=1/1,framerate=30/1"),
                ("mid", "video/x-raw,format=YUY2,width=800,height=600,pixel-aspect-ratio=1/1,framerate=20/1"),
                ("high", "video/x-raw,format=YUY2,width=1920,height=1080,pixel-aspect-ratio=1/1,framerate=5/1"),
                ])
        self.logger.warning(
            "Unknown USB video device {:04x}:{:04x}, using fallback pipeline."
            .format(*match))
        return ("mid", [
            ("low", "image/jpeg,width=640,height=480,framerate=30/1"),
            ("mid", "image/jpeg,width=1280,height=720,framerate=30/1"),
            ("high", "image/jpeg,width=1920,height=1080,framerate=30/1"),
            ])

    def select_caps(self, hint=None):
        default, variants = self.get_qualities()
        variant = hint if hint else default
        for name, caps in variants:
            if name == variant:
                return caps
        raise InvalidConfigError(
            f"Unknown video format {variant} for device {self.video.vendor_id:04x}:{self.video.model_id:04x}"  # pylint: disable=line-too-long
        )

    def get_pipeline(self, path, caps, controls=None):
        match = (self.video.vendor_id, self.video.model_id)
        if match == (0x046d, 0x082d):
            controls = controls or "focus_auto=1"
            inner = "h264parse"
        elif match == (0x046d, 0x0892):
            controls = controls or "focus_auto=1"
            inner = None
        elif match == (0x046d, 0x08e5):
            controls = controls or "focus_auto=1"
            inner = None
        elif match == (0x1224, 0x2825): # LogiLink UA0371
            inner = None  # just forward the jpeg frames
        elif match == (0x05a3, 0x9331): # WansView Webcam 102
            inner = "h264parse"
        elif match == (0x534d, 0x2109):
            inner = None  # just forward the jpeg frames
        elif match == (0x1d6c, 0x0103):
            controls = controls or "focus_auto=1"
            inner = "h264parse"
        elif match == (0x0c54, 0x636b):
            controls = controls or "focus_auto=1"
            inner = None  # just forward the jpeg frames
        elif match == (0x0c54, 0x636d):
            controls = controls or "focus_auto=1"
            inner = None  # just forward the jpeg frames
        else: # fallback pipeline
            inner = None  # just forward the jpeg frames

        pipeline = f"v4l2src device={path} "
        if controls:
            pipeline += f"extra-controls=c,{controls} "
        pipeline += f"! {caps} "
        if inner:
            pipeline += f"! {inner} "
        pipeline += "! matroskamux streamable=true ! fdsink"
        return pipeline

    @Driver.check_active
    def stream(self, caps_hint=None, controls=None):
        caps = self.select_caps(caps_hint)
        pipeline = self.get_pipeline(self.video.path, caps, controls)

        tx_cmd = self.video.command_prefix + ["gst-launch-1.0", "-q"]
        tx_cmd += pipeline.split()
        rx_cmd = ["gst-launch-1.0", "playbin3", "buffer-duration=0", "uri=fd://0"]

        tx = subprocess.Popen(
            tx_cmd,
            stdin=subprocess.DEVNULL,
            stdout=subprocess.PIPE,
        )
        rx = subprocess.Popen(
            rx_cmd,
            stdin=tx.stdout,
            stdout=subprocess.DEVNULL,
        )

        # wait until one subprocess has terminated
        while True:
            try:
                tx.wait(timeout=0.1)
                break
            except subprocess.TimeoutExpired:
                pass
            try:
                rx.wait(timeout=0.1)
                break
            except subprocess.TimeoutExpired:
                pass

        rx.terminate()
        tx.terminate()

        rx.communicate()
        tx.communicate()



"""tested with Siglent SPD3303X-E, and should be compatible with all SPD3000X series modules"""

import warnings

import vxi11


def power_set(host, port, index, value):
    warnings.warn(
        "siglent power backend uses vxi11 module using deprecated xdrlib module, see https://github.com/labgrid-project/labgrid/issues/1507",
        DeprecationWarning,
    )

    assert port is None
    index = int(index)
    assert 1 <= index <= 2
    value = "ON" if value else "OFF"
    psu = vxi11.Instrument(host)
    psu.write(f"OUTPUT CH{index},{value}")


def power_get(host, port, index):
    warnings.warn(
        "siglent power backend uses vxi11 module using deprecated xdrlib module, see https://github.com/labgrid-project/labgrid/issues/1507",
        DeprecationWarning,
    )

    assert port is None
    index = int(index)
    assert 1 <= index <= 2
    psu = vxi11.Instrument(host)
    state = psu.ask("SYSTEM:STATUS?")
    state = int(state, 16)
    bitmask = 1 << (index + 3)
    return bool(state & bitmask)



from labgrid.resource.mqtt import TasmotaPowerPort
from labgrid.driver.mqtt import TasmotaPowerDriver

import pytest

pytest.importorskip("paho.mqtt.client")


def test_tasmota_resource(target, mocker):
    mocker.patch("paho.mqtt.client.Client.connect", return_value=None)
    mocker.patch("paho.mqtt.client.Client.loop_start", return_value=None)
    pw_set = mocker.patch("paho.mqtt.client.Client.username_pw_set", return_value=None)
    TasmotaPowerPort(target, name=None, host="localhost", avail_topic="test", power_topic="test", status_topic="test")
    pw_set.assert_called_once_with(None, None)


def test_tasmota_driver(target, mocker):
    mocker.patch("paho.mqtt.client.Client.connect", return_value=None)
    mocker.patch("paho.mqtt.client.Client.loop_start", return_value=None)
    pw_set = mocker.patch("paho.mqtt.client.Client.username_pw_set", return_value=None)
    res = TasmotaPowerPort(
        target,
        name=None,
        host="localhost",
        avail_topic="test",
        power_topic="test",
        status_topic="test",
        username="foo",
        password="bar",
    )
    res.manager._available.add("test")
    driver = TasmotaPowerDriver(target, name=None)
    target.activate(driver)
    pw_set.assert_called_once_with("foo", "bar")



import contextlib
import errno
import functools
import logging
import os
import random
import socket
import string
import subprocess
import sys
import tempfile

from . import get_free_port
from .agent import py2s, s2py


class NSSocket(socket.socket):
    """
    Proxy class for sockets from a different netns.

    There are certain socket operations that must be called in the network
    namespace to work properly. Instead of making custom wrappers for thes in
    the NetNamespace object, the NSSocket class derives from `socket.socket`,
    but transparently redirects applicable socket calls into the namespace.
    This relies on the namespace tracking a mirror socket file descriptor for
    each created NSSocket, and when NSSocket forwards calls into the network
    namespace, it performs the operation on the mirror socket. This works
    because multiple file descriptors can refer to the same actual socket, even
    in different namespaces.
    """

    # __init__ must socket.socket() constructor, due to the way python uses it
    # internally. _attach_remote_sock() can be used after the socket is created
    # to attach the remote socket ID
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__sockid = None
        self.__netns = None

    def _attach_remote_sock(self, sockid, netns):
        self.__sockid = sockid
        self.__netns = netns

        if sockid:
            self.__forward_function("connect")
            self.__forward_function("connect_ex")
            self.__forward_function("bind")

        return self

    def __close_remote_sock(self):
        if self.__sockid:
            self.__netns.socket_close(self.__sockid)
            self.__sockid = None

    def __forward_function(self, name):
        def wrap(*args, **kwargs):
            return self.__netns_call(name, args, kwargs)

        setattr(self, name, wrap)

    def __netns_call(self, func, args, kwargs):
        if not self.__sockid:
            raise OSError(errno.EBADF, os.strerror(errno.EBADF))

        err, ret = self.__netns.socket_call(self.__sockid, func, py2s(args), py2s(kwargs))
        if err != 0:
            raise OSError(*err)
        return s2py(ret)

    def close(self):
        self.__close_remote_sock()
        return super().close()

    def detach(self):
        # NOTE: The detached file descriptor no longer forwards API to the
        # namespace
        self.__close_remote_sock()
        return super().detach()

    def dup(self):
        if self.__sockid:
            ret, fd = self.__netns.socket_dup(self.__sockid)
            if "error" in ret:
                raise OSError(*ret["error"])

            s = self.__class__(fileno=fd)._attach_remote_sock(ret["id"], self.__netns)
            s.settimeout(self.gettimeout())
            return s
        else:
            return super().dup()


class NetNamespace:
    @classmethod
    def create(cls, agentwrapper, mac_address=None):
        netns = agentwrapper.load("netns")
        netns.unshare()
        netns.create_tun(address=mac_address)
        return cls(netns)

    def __init__(self, agent):
        self.logger = logging.getLogger(self.__class__.__name__)
        self._agent = agent

    @functools.cached_property
    def prefix(self):
        return self._agent.get_prefix()

    @functools.cached_property
    def intf(self):
        return self._agent.get_intf()

    def ifindex(self):
        links = self._agent.get_links()
        for intf in links:
            if intf["ifname"] == self.intf:
                return intf["ifindex"]
        raise KeyError(f"Interface {self.intf} not found")

    def get_links(self):
        return self._agent.get_links()

    def _get_cmd(self, command):
        if isinstance(command, str):
            return self.prefix + ["--wd=" + os.getcwd(), "--", "/bin/sh", "-c", command]
        return self.prefix + ["--wd=" + os.getcwd(), "--"] + command

    def run(self, command, **kwargs):
        cmd = self._get_cmd(command)
        self.logger.debug("Running %s", cmd)
        return subprocess.run(cmd, **kwargs)

    def Popen(self, command, **kwargs):
        cmd = self._get_cmd(command)
        self.logger.debug("Popen %s", cmd)
        return subprocess.Popen(cmd, **kwargs)

    @contextlib.contextmanager
    def _create_script(self, script):
        with tempfile.NamedTemporaryFile("w") as s:
            s.write(script)
            s.flush()

            yield [sys.executable, s.name]

    def run_script(self, script, script_args=[], **kwargs):
        with self._create_script(script) as command:
            return self.run(command + script_args, **kwargs)

    @contextlib.contextmanager
    def Popen_script(self, script, script_args=[], **kwargs):
        with self._create_script(script) as command:
            with self.Popen(command + script_args, **kwargs) as p:
                yield p

    def socket(self, *args, **kwargs):
        ret, fd = self._agent.create_socket(*args, **kwargs)
        if "error" in ret:
            raise OSError(*ret["error"])
        return NSSocket(fileno=fd)._attach_remote_sock(ret["id"], self._agent)

    def getaddrinfo(self, *args, **kwargs):
        err, result = self._agent.getaddrinfo(*args, **kwargs)
        if err:
            raise OSError(*err)
        return result

    @contextlib.contextmanager
    def socks_proxy(self):
        port = get_free_port()
        password = "".join(random.choice(string.ascii_letters + string.digits) for i in range(10))

        with contextlib.ExitStack() as stack:
            proxy = stack.enter_context(
                self.Popen(
                    ["microsocks", "-u", "labgrid", "-P", password, "-i", "127.0.0.1"],
                )
            )
            stack.callback(lambda: proxy.terminate())
            s = stack.enter_context(socket.socket())
            s.bind(("127.0.0.1", port))
            s.listen()
            socat = stack.enter_context(
                self.Popen(
                    ["socat", f"ACCEPT-FD:{s.fileno()},fork", "TCP4:127.0.0.1:1080"],
                    pass_fds=(s.fileno(),),
                )
            )
            stack.callback(lambda: socat.terminate())
            stack.callback(lambda: print("A"))
            s.close()
            yield (port, password)



import os
import subprocess
import pytest

from ..exceptions import NoResourceFoundError, NoDriverFoundError
from ..remote.client import UserError
from ..resource.remote import RemotePlace
from ..util.ssh import sshmanager
from ..logging import DEFAULT_FORMAT
from .hooks import LABGRID_ENV_KEY

# pylint: disable=redefined-outer-name


def pytest_addoption(parser):
    group = parser.getgroup('labgrid')
    group.addoption(
        '--lg-env',
        action='store',
        dest='lg_env',
        help='labgrid environment config file.')
    group.addoption(
        '--lg-coordinator',
        action='store',
        dest='lg_coordinator',
        metavar='COORDINATOR_ADDRESS',
        help='labgrid coordinator HOST[:PORT].')
    group.addoption(
        '--lg-log',
        action='store',
        dest='lg_log',
        metavar='path to logfiles',
        nargs='?',
        const=".",
        help='path to store logfiles')
    group.addoption(
        '--lg-colored-steps',
        action='store_true',
        dest='lg_colored_steps',
        help='colored step reporter')
    group.addoption(
        '--lg-initial-state',
        action='store',
        dest='lg_initial_state',
        metavar='STATE_NAME',
        help='set the strategy\'s initial state (during development)')

    # We would like to use a default value hook for log_format in the logging plugin,
    # similar to the approach below:
    # https://docs.pytest.org/en/latest/how-to/writing_hook_functions.html#using-hooks-in-pytest-addoption
    parser.addini("log_format", default=DEFAULT_FORMAT, help="Default value for log_format (overwritten by labgrid)")


@pytest.fixture(scope="session")
def env(request, record_testsuite_property):
    """Return the environment configured in the supplied configuration file.
    It contains the targets contained in the configuration file.
    """
    env = request.config.stash[LABGRID_ENV_KEY]

    if not env:
        pytest.skip("missing environment config (use --lg-env)")

    record_testsuite_property('ENV_CONFIG', env.config_file)
    targets = list(env.config.get_targets().keys())
    record_testsuite_property('TARGETS', targets)

    for target_name in targets:
        try:
            target = env.get_target(target_name)
        except UserError as e:
            pytest.exit(e)
        try:
            remote_place = target.get_resource(RemotePlace, wait_avail=False)
            remote_name = remote_place.name
            record_testsuite_property(
                f'TARGET_{target_name.upper()}_REMOTE', remote_name)
        except NoResourceFoundError:
            pass

    for name, path in env.config.get_paths().items():
        record_testsuite_property(f'PATH_{name.upper()}', path)
        try:
            sha = subprocess.check_output(
                "git rev-parse HEAD".split(), cwd=path, stderr=subprocess.DEVNULL)
        except subprocess.CalledProcessError:
            continue
        except FileNotFoundError:
            continue
        record_testsuite_property(
            f'PATH_{name.upper()}_GIT_COMMIT',
            sha.decode("utf-8").strip("\n"))

    for name, image in env.config.get_images().items():
        record_testsuite_property(
            f'IMAGE_{name.upper()}', image)
        try:
            sha = subprocess.check_output(
                "git rev-parse HEAD".split(), cwd=os.path.dirname(image),
                stderr=subprocess.DEVNULL)
        except subprocess.CalledProcessError:
            continue
        except FileNotFoundError:
            continue
        record_testsuite_property(
            f'IMAGE_{name.upper()}_GIT_COMMIT',
            sha.decode("utf-8").strip("\n"))

    yield env
    env.cleanup()
    sshmanager.close_all()


@pytest.fixture(scope="session")
def target(env):
    """Return the default target "main" configured in the supplied
    configuration file."""
    target = env.get_target()
    if target is None:
        raise UserError("Using target fixture without 'main' target in config")

    return target


@pytest.fixture(scope="session")
def strategy(request, target):
    """Return the Strategy of the default target "main" configured in the
    supplied configuration file."""
    try:
        strategy = target.get_driver("Strategy")
    except NoDriverFoundError as e:
        pytest.exit(e)

    state = request.config.option.lg_initial_state
    if state is not None:
        strategy.force(state)

    return strategy



def test_usb_storage(target):
    r = target.get_resource("USBMassStorage")
    assert r.path is not None


def test_usb_storage_size(target):
    d = target.get_driver("USBStorageDriver")
    assert d.get_size() > 0



import string

import pexpect


class PtxExpect(pexpect.spawn):
    """labgrid Wrapper of the pexpect module.

    This class provides pexpect functionality for the ConsoleProtocol classes.
    driver: ConsoleProtocol object to be passed in
    """

    def __init__(self, driver):
        "Initializes a pexpect spawn instance with the required configuration"
        self.driver = driver
        self.linesep = b"\n"
        pexpect.spawn.__init__(self, None, maxread=1)

    def send(self, s):
        "Write to underlying transport, return number of bytes written"
        s = self._coerce_send_string(s)
        self._log(s, 'send')

        b = s
        return self.driver.write(b)

    def sendcontrol(self, char):
        """Send control character to underlying transport, e.g. Ctrl-C or Ctrl-Z."""
        char = char.lower()
        try:
            ord_ = string.ascii_lowercase.index(char) + 1
            self.send(bytes([ord_]))
        except ValueError:
            raise NotImplementedError(f"Sending control character {char} is not supported yet")

    def read_nonblocking(self, size=1, timeout=-1):
        """Pexpects needs a nonblocking read function, simply use pyserial with
        a timeout of 0."""
        assert timeout is not None
        if timeout == -1:
            timeout = self.timeout
        return self.driver.read(size=size, timeout=timeout)



from importlib import import_module
import attr

from ..factory import target_factory
from .common import Driver


@target_factory.reg_driver
@attr.s(eq=False)
class ModbusRTUDriver(Driver):
    bindings = {"resource": "ModbusRTU"}

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self._modbus = import_module('minimalmodbus')
        self.instrument = None

    def on_activate(self):
        self.instrument = self._modbus.Instrument(
            self.resource.port,
            self.resource.address,
            debug=False)
        self.instrument.serial.baudrate = self.resource.speed
        self.instrument.serial.timeout = self.resource.timeout

        self.instrument.mode = self._modbus.MODE_RTU
        self.instrument.clear_buffers_before_each_transaction = True

    def on_deactivate(self):
        self.instrument = None

    def read_register(self, *args, **kwargs):
        return self.instrument.read_register(*args, **kwargs)

    def write_register(self, *args, **kwargs):
        return self.instrument.write_register(*args, **kwargs)

    def read_registers(self, *args, **kwargs):
        return self.instrument.read_registers(*args, **kwargs)

    def write_registers(self, *args, **kwargs):
        return self.instrument.write_registers(*args, **kwargs)

    def read_bit(self, *args, **kwargs):
        return self.instrument.read_bit(*args, **kwargs)

    def write_bit(self, *args, **kwargs):
        return self.instrument.write_bit(*args, **kwargs)

    def read_string(self, *args, **kwargs):
        return self.instrument.read_string(*args, **kwargs)

    def write_string(self, *args, **kwargs):
        return self.instrument.write_string(*args, **kwargs)



import attr

from ..factory import target_factory
from .common import NetworkResource, Resource


@target_factory.reg_resource
@attr.s(eq=False)
class YKUSHPowerPort(Resource):
    """This resource describes a YEPKIT YKUSH switchable USB hub.

    Args:
        serial (str): serial of the YKUSH device
        index (str): port index"""

    serial = attr.ib(validator=attr.validators.instance_of(str))
    index = attr.ib(validator=attr.validators.instance_of(str), converter=str)


@target_factory.reg_resource
@attr.s(eq=False)
class NetworkYKUSHPowerPort(NetworkResource):
    """ "This resource describes a remote YEPKIT YKUSH switchable USB hub.

    Args:
        serial (str): serial of the YKUSH device
        index (str): port index"""

    serial = attr.ib(validator=attr.validators.instance_of(str))
    index = attr.ib(validator=attr.validators.instance_of(str), converter=str)



import asyncio

from labgrid.resource import SNMPEthernetPort


def test_instance(target):
    # SNMPEthernetPort should be called with a running event loop
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    try:
        s = SNMPEthernetPort(target, "port-1", switch="dummy-switch", interface="1")
        assert isinstance(s, SNMPEthernetPort)
    finally:
        loop.close()



"""
This module contains the custom YAML load and dump functions and associated
loader and dumper
"""

import warnings
from collections import OrderedDict, UserString
from string import Template

import yaml


class Loader(yaml.SafeLoader):
    pass


class Dumper(yaml.SafeDumper):
    pass


def _check_duplicate_dict_keys(loader, node):
    seen_keys = []
    for key_node, _ in node.value:
        key = loader.construct_scalar(key_node)
        if key in seen_keys:
            warnings.warn(
                f"{loader.name}: previous entry with duplicate YAML dictionary key '{key}' overwritten", UserWarning
            )
        seen_keys.append(key)


def _dict_constructor(loader, node):
    _check_duplicate_dict_keys(loader, node)
    return OrderedDict(loader.construct_pairs(node))


Loader.add_constructor(yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, _dict_constructor)
Loader.add_constructor(
    "tag:yaml.org,2002:python/tuple",
    yaml.constructor.FullConstructor.construct_python_tuple,
)


def _dict_representer(dumper, data):
    return dumper.represent_dict(data.items())


Dumper.add_representer(OrderedDict, _dict_representer)


def _str_constructor(loader, node):
    # store location of multiline string
    if node.style != "|":
        return loader.construct_scalar(node)
    obj = UserString(loader.construct_scalar(node))
    obj.start_mark = node.start_mark
    obj.end_mark = node.end_mark
    return obj


Loader.add_constructor(yaml.resolver.BaseResolver.DEFAULT_SCALAR_TAG, _str_constructor)


def _template_constructor(loader, node):
    return Template(loader.construct_scalar(node))


Loader.add_constructor("!template", _template_constructor)


def load(stream):
    """
    Wrapper for yaml load function with custom loader.
    """
    return yaml.load(stream, Loader=Loader)


def dump(data, stream=None, **kwargs):
    """
    Wrapper for yaml dump function with custom dumper.
    """
    kwargs.pop("Dumper", None)
    return yaml.dump(data, stream, Dumper=Dumper, **kwargs)


def resolve_templates(data, mapping):
    """
    Iterate recursively over data and call substitute(mapping) on all
    Templates.
    """
    if isinstance(data, list):
        items = enumerate(data)
    elif isinstance(data, dict):
        items = data.items()
    else:
        raise TypeError(f"Expected list or dict, got {type(data)}")

    for k, val in items:
        if isinstance(val, Template):
            try:
                data[k] = val.substitute(mapping)
            except ValueError as error:
                raise ValueError(f"Invalid template string '{val.template}'") from error

        elif isinstance(val, (list, dict)):
            resolve_templates(val, mapping)



from ..exception import ExecutionError
from ...util.helper import processwrapper

OID = ".1.3.6.1.4.1.318.1.1.4.4.2.1.3"


def _snmp_get(host, oid):
    out = processwrapper.check_output(f"snmpget -v1 -c private -O qn {host} {oid}".split()).decode("ascii")
    out_oid, value = out.strip().split(" ", 1)
    assert oid == out_oid
    if value == "1":
        return True
    if value == "2":
        return False

    raise ExecutionError("failed to get SNMP value")


def _snmp_set(host, oid, value):
    try:
        processwrapper.check_output(f"snmpset -v1 -c private {host} {oid} {value}".split())
    except Exception as e:
        raise ExecutionError("failed to set SNMP value") from e


def power_set(host, port, index, value):
    assert port is None

    index = int(index)
    value = 1 if value else 2
    assert 1 <= index <= 8

    _snmp_set(host, f"{OID}.{index}", f"int {value}")


def power_get(host, port, index):
    assert port is None

    index = int(index)
    assert 1 <= index <= 8

    return _snmp_get(host, f"{OID}.{index}")



import abc
import atexit
import logging
import sys
from time import monotonic, sleep

import attr

from .binding import BindingError, BindingState
from .driver import Driver
from .exceptions import NoSupplierFoundError, NoDriverFoundError, NoResourceFoundError, NoStrategyFoundError
from .resource import Resource
from .strategy import Strategy
from .util import Timeout
from .factory import target_factory


@attr.s(eq=False)
class Target:
    name = attr.ib(validator=attr.validators.instance_of(str))
    env = attr.ib(default=None)

    def __attrs_post_init__(self):
        self.log = logging.getLogger(f"target({self.name})")
        self.resources = []
        self.drivers = []
        self.last_update = 0.0
        # This should really be an argument for Drivers, but currently attrs
        # doesn't support keyword only agruments, so we can't add an optional
        # argument at the BindingMixin level.
        # https://github.com/python-attrs/attrs/issues/106
        self._binding_map = {}
        self._lookup_table = {
            Strategy.__name__: Strategy,
        }
        atexit.register(self._atexit_cleanup)

    def interact(self, msg):
        if self.env:
            self.env.interact(f"{self.name}: {msg}")
        else:
            input(msg)

    def update_resources(self):
        """
        Iterate over this target's resources, deactivate any active but
        unavailable resources and also deactivate any drivers using them.
        This ensures a consistent binding states for this target.
        """
        if (monotonic() - self.last_update) < 0.1:
            return
        self.last_update = monotonic()
        for resource in self.resources:
            resource.poll()
            if not resource.avail and resource.state is BindingState.active:
                deactivated = self.deactivate(resource)
                deactivated.remove(resource)
                if deactivated:
                    self.log.info("deactivating unavailable resource %s used by %s",
                        resource.display_name,
                        ", ".join(d.display_name for d in deactivated)
                    )
                else:
                    self.log.debug("deactivating unavailable resource %s (unused)", resource.display_name)  # pylint: disable=line-too-long

    def await_resources(self, resources, timeout=None, avail=True):
        """
        Poll the given resources and wait until they are (un-)available.

        Args:
            resources (List): the resources to poll
            timeout (float): optional timeout
            avail (bool): optionally wait until the resources are unavailable with avail=False
        """
        self.update_resources()

        waiting = set(r for r in resources if r.avail != avail)
        static = set(r for r in waiting if r.get_managed_parent() is None)
        if static:
            raise NoResourceFoundError(
                f"Static resources are not {'available' if avail else 'unavailable'}: {static}"
            )

        if not waiting:
            return

        if timeout is None:
            timeout = Timeout(max(resource.get_managed_parent().timeout for resource in waiting))
        else:
            timeout = Timeout(timeout)

        while waiting and not timeout.expired:
            waiting = set(r for r in waiting if r.avail != avail)
            for r in waiting:
                r.poll()
            if not any(r for r in waiting if r.avail == avail):
                # sleep if no progress
                sleep(0.5)

        if waiting:
            raise NoResourceFoundError(
                f"Not all resources are {'available' if avail else 'unavailable'}: {waiting}",
                filter=waiting
            )

        self.update_resources()

    def get_resource(self, cls, *, name=None, wait_avail=True):
        """
        Helper function to get a resource of the target.
        Returns the first valid resource found, otherwise a
        NoResourceFoundError is raised.

        Arguments:
        cls -- resource-class to return as a resource
        name -- optional name to use as a filter
        wait_avail -- wait for the resource to become available (default True)
        """
        found = []
        other_names = []
        default = None
        if isinstance(cls, str):
            cls = target_factory.class_from_string(cls)

        for res in self.resources:
            if not isinstance(res, cls):
                continue
            if res.name == "default":
                default = res
            if name and res.name != name:
                other_names.append(res.name)
                continue
            found.append(res)

        # if no explicit resource name is requested and a "default" resource was saved and
        # multiple resources were found, use the default resource
        if not name and default and len(found) != 1:
            found = [default]

        if not found:
            name_msg = f" named '{name}'" if name else ""
            if other_names:
                raise NoResourceFoundError(
                    f"no {cls.__name__} resource{name_msg} found in {self}, matching resources with other names: {other_names}"  # pylint: disable=line-too-long
                )

            raise NoResourceFoundError(
                f"no {cls.__name__} resource{name_msg} found in {self}"
            )
        elif len(found) > 1:
            raise NoResourceFoundError(
                f"multiple resources matching {cls.__name__} found in {self}", found=found
            )
        if wait_avail:
            self.await_resources(found)
        return found[0]

    def _get_driver(self, cls, *, name=None, resource=None, activate=True, active=False):
        assert not (activate is True and active is True)

        found = []
        other_names = []
        if isinstance(cls, str):
            cls = target_factory.class_from_string(cls)

        for drv in self.drivers:
            if not isinstance(drv, cls):
                continue
            if resource and resource not in drv.get_bound_resources():
                continue
            if name and drv.name != name:
                other_names.append(drv.name)
                continue
            if active and drv.state != BindingState.active:
                continue
            found.append(drv)
        if not found:
            name_msg = f" named '{name}'" if name else ""
            if other_names:
                raise NoDriverFoundError(
                    "no {active}{cls} driver{name} found in {target}, matching resources with other names: {other_names}".format(  # pylint: disable=line-too-long
                        active="active " if active else "", cls=cls.__name__, name=name_msg,
                        target=self, other_names=other_names)
                )

            raise NoDriverFoundError(
                f"no {'active ' if active else ''}{cls.__name__} driver{name_msg} found in {self}"
            )
        elif len(found) > 1:
            prio_last = -255
            prio_found = []
            for drv in found:
                prio = drv.get_priority(cls)
                if prio > prio_last:
                    prio_found = []
                    prio_found.append(drv)
                    prio_last = prio
                elif prio == prio_last:
                    prio_found.append(drv)

            if len(prio_found) == 1:
                found = prio_found
            else:
                raise NoDriverFoundError(
                    "multiple {active}drivers matching {cls} found in {target} with the same priorities".format(  # pylint: disable=line-too-long
                        active="active " if active else "", cls=cls.__name__, target=self)
                )
        if activate:
            self.activate(found[0])
        return found[0]

    def get_active_driver(self, cls, *, name=None, resource=None):
        """
        Helper function to get the active driver of the target.
        Returns the active driver found, otherwise None.

        Arguments:
        cls -- driver-class to return as a resource
        name -- optional name to use as a filter
        resource -- optional resource to use as a filter
        """
        return self._get_driver(cls, name=name, resource=resource, activate=False, active=True)

    def get_driver(self, cls, *, name=None, resource=None, activate=True):
        """
        Helper function to get a driver of the target.
        Returns the first valid driver found, otherwise None.

        Arguments:
        cls -- driver-class to return as a resource
        name -- optional name to use as a filter
        resource -- optional resource to use as a filter
        activate -- activate the driver (default True)
        """
        return self._get_driver(cls, name=name, resource=resource, activate=activate)

    def get_strategy(self):
        """
        Helper function to get the strategy of the target.

        Returns the Strategy, if exactly one exists and raises a
        NoStrategyFoundError otherwise.
        """
        found = []
        for drv in self.drivers:
            if not isinstance(drv, Strategy):
                continue
            found.append(drv)
        if not found:
            raise NoStrategyFoundError(f"no Strategy found in {self}")
        elif len(found) > 1:
            raise NoStrategyFoundError(f"multiple Strategies found in {self}")
        return found[0]

    def __getitem__(self, key):
        """
        Syntactic sugar to access drivers by class (optionally filtered by
        name).

        >>> from labgrid import Target
        >>> from labgrid.driver.fake import FakeConsoleDriver
        >>>
        >>> target = Target('main')
        >>> console = FakeConsoleDriver(target, 'console')
        >>> target.activate(console)
        >>> target[FakeConsoleDriver]
        FakeConsoleDriver(target=Target(name='main', env=None), name='console', state=<BindingState.active: 2>, txdelay=0.0, txchunk=1)
        >>> target[FakeConsoleDriver, 'console']
        FakeConsoleDriver(target=Target(name='main', env=None), name='console', state=<BindingState.active: 2>, txdelay=0.0, txchunk=1)
        """
        name = None
        if not isinstance(key, tuple):
            cls = key
        elif len(key) == 2:
            cls, name = key
        if isinstance(cls, str):
            cls = target_factory.class_from_string(cls)
        if not issubclass(cls, (Driver, abc.ABC)): # all Protocols derive from ABC
            raise NoDriverFoundError(
                f"invalid driver class {cls.__name__}"
            )

        return self.get_active_driver(cls, name=name)

    def set_binding_map(self, mapping):
        """
        Configure the binding name mapping for the next driver only.
        """
        self._binding_map = mapping

    def bind_resource(self, resource):
        """
        Bind the resource to this target.
        """
        if resource.state is not BindingState.idle:
            raise BindingError(
                f"{resource} is not in state {BindingState.idle}"
            )

        # consistency check
        assert isinstance(resource, Resource)
        assert not resource.bindings
        assert resource not in self.resources
        assert resource.target is None

        # update state
        self.resources.append(resource)
        # update lookup table
        self._lookup_table[resource.__class__.__name__] = resource.__class__
        resource.target = self
        resource.state = BindingState.bound

    def bind_driver(self, client):
        """
        Bind the driver to all suppliers (resources and other drivers).

        Currently, we only support binding all suppliers at once.
        """
        if client.state is not BindingState.idle:
            raise BindingError(
                f"{client} is not in state {BindingState.idle}"
            )

        # consistency check
        assert isinstance(client, Driver)
        assert client not in self.drivers
        assert client.target is None

        mapping = self._binding_map
        self._binding_map = {}

        # locate suppliers
        bound_suppliers = []
        bound_req_pairs = set()
        for name, requirements in client.bindings.items():
            explicit = False
            if isinstance(requirements, Driver.NamedBinding):
                requirements = requirements.value
                explicit = True
            supplier_name = mapping.pop(name, None)
            if explicit and supplier_name is None:
                raise BindingError(
                    f"supplier for {name} ({requirements}) of {client} in {self} requires an explicit name"  # pylint: disable=line-too-long
                )
            # use sets even for a single requirement and make a local copy
            if not isinstance(requirements, set):
                requirements = {requirements}
            else:
                requirements = requirements.copy()
            # None indicates that the binding is optional
            optional = None in requirements
            requirements.discard(None)

            errors = []
            suppliers = []
            for requirement in requirements:
                # convert class name string to classes
                if isinstance(requirement, str):
                    requirement = target_factory.class_from_string(requirement)
                try:
                    if issubclass(requirement, Resource):
                        suppliers.append(
                            self.get_resource(requirement, name=supplier_name, wait_avail=False),
                        )
                    elif issubclass(requirement, (Driver, abc.ABC)): # all Protocols derive from ABC
                        suppliers.append(
                            self.get_driver(requirement, name=supplier_name, activate=False),
                        )
                    else:
                        raise NoSupplierFoundError(f"invalid binding type {requirement}")
                except NoSupplierFoundError as e:
                    errors.append(e)
            if not suppliers:
                client_name = client.name or client.__class__.__name__
                if optional:
                    supplier = None
                elif len(errors) == 1:
                    err = errors[0]
                    err_cls = type(err)
                    raise err_cls(f"binding {client_name} failed: {err}") from err
                else:
                    raise NoSupplierFoundError(
                        f"binding {client_name} failed: no supplier matching {requirements} found in {self} (errors: {errors})"
                    )
            elif len(suppliers) > 1:
                raise NoSupplierFoundError(f"conflicting suppliers matching {requirements} found in target {self}")  # pylint: disable=line-too-long
            else:
                supplier = suppliers[0]
            if supplier is not None and (requirement, supplier) in bound_req_pairs:
                raise BindingError(
                    f"duplicate bindings of {supplier} to {name} for {requirement} found in target {self}"  # pylint: disable=line-too-long
                )
            bound_req_pairs.add((requirement, supplier))
            setattr(client, name, supplier)
            if supplier is not None:
                bound_suppliers.append(supplier)

        # consistency checks
        for supplier in bound_suppliers:
            assert supplier.target is self
            assert client not in supplier.clients
            assert supplier not in client.suppliers

        # make sure drivers consume all given bindings
        if mapping and not isinstance(client, Strategy):
            raise BindingError(
                f"{client} got unexpected bindings: {list(mapping.keys())}"
            )

        # update relationship in both directions
        self.drivers.append(client)
        # update lookup table
        cls = client.__class__
        self._lookup_table[cls.__name__] = cls
        for c in cls.mro():
            if abc.ABC in c.mro():
                self._lookup_table[c.__name__] = c

        client.target = self
        for supplier in bound_suppliers:
            supplier.clients.add(client)
            client.suppliers.add(supplier)
            client.on_supplier_bound(supplier)
            supplier.on_client_bound(client)
        client.state = BindingState.bound

    def bind(self, bindable):
        if isinstance(bindable, Resource):
            return self.bind_resource(bindable)
        if isinstance(bindable, Driver):
            return self.bind_driver(bindable)

        raise BindingError(f"object {bindable} is not bindable")

    def activate(self, client, name=None):
        """
        Activate the client by activating all bound suppliers. This may require
        deactivating other clients.
        """
        # don't activate strategies, they usually have conflicting bindings
        if isinstance(client, Strategy):
            return

        if isinstance(client, str):
            cls = target_factory.class_from_string(client)
            client = self._get_driver(cls, name=name, activate=False, active=False)

        assert client is not None

        if client.state is BindingState.active:
            return  # nothing to do

        if client.state is not BindingState.bound:
            raise BindingError(
                f"{client} is not in state {BindingState.bound}"
            )

        # consistency check
        assert client in self.resources or client in self.drivers

        # wait until resources are available
        resources = [resource for resource in client.suppliers if isinstance(resource, Resource)]
        self.await_resources(resources)

        # activate recursively and resolve conflicts
        for supplier in client.suppliers:
            if supplier.state is not BindingState.active:
                self.activate(supplier)
            supplier.resolve_conflicts(client)

        # update state
        client.on_activate()
        client.state = BindingState.active

    def deactivate(self, client, name=None):
        """
        Recursively deactivate the client's clients and itself.

        This is needed to ensure that no client has an inactive supplier.

        Returns the list of all objects that were deactivated
        """
        if isinstance(client, str):
            cls = target_factory.class_from_string(client)
            client = self._get_driver(cls, name=name, activate=False, active=True)

        assert client is not None

        if client.state is BindingState.bound:
            return [] # nothing to do

        if client.state is not BindingState.active:
            raise BindingError(
                f"{client} is not in state {BindingState.active}"
            )

        # consistency check
        assert client in self.resources or client in self.drivers

        deactivated = [client]

        for cli in client.clients:
            deactivated.extend(self.deactivate(cli))

        # update state
        client.on_deactivate()
        client.state = BindingState.bound
        return deactivated

    def deactivate_all_drivers(self):
        """Deactivates all drivers in reversed order they were activated"""
        for drv in reversed(self.drivers):
            self.deactivate(drv)

    def _atexit_cleanup(self):
        try:
            self.cleanup()
        except Exception as e:
            print("An exception occured during cleanup, call the cleanup() "
                  "method on targets yourself to handle exceptions explictly.", file=sys.stderr)
            print(f"Error: {e}", file=sys.stderr)
            import traceback
            traceback.print_exc(file=sys.stderr)

    def export(self):
        """
        Export information from drivers.

        All drivers are deactivated before being exported, unless their
        skip_deactivate_on_export property is true.

        The Strategy can decide for which driver the export method is called and
        with which name. Otherwise, all drivers are exported.
        """
        try:
            name_map = self.get_strategy().prepare_export()
            selection = set(name_map.keys())
        except NoStrategyFoundError:
            name_map = {}
            selection = set(driver for driver in self.drivers if not isinstance(driver, Strategy))

        assert len(name_map) == len(set(name_map.values())), "duplicate export name"

        # drivers may need to be deactivated for export to avoid conflicts
        for drv in reversed(self.drivers):
            if not drv.skip_deactivate_on_export:
                self.deactivate(drv)

        export_vars = {}
        for driver in selection:
            name = name_map.get(driver)
            if not name:
                name = driver.get_export_name()
            for k, v in driver.get_export_vars().items():
                assert isinstance(k, str), f"key {k} from {driver} is not a string"
                assert isinstance(v, str), f"value {v} for key {k} from {driver} is not a string"
                export_vars[f"LG__{name}_{k}".upper()] = v
        return export_vars


    def cleanup(self):
        """Clean up connected drivers and resources in reversed order"""
        self.deactivate_all_drivers()
        for res in reversed(self.resources):
            self.deactivate(res)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.cleanup()



import copy
import os
import attr

from ..factory import target_factory
from .common import NetworkResource, ManagedResource, ResourceManager


@attr.s(eq=False)
class RemotePlaceManager(ResourceManager):
    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self.url = None
        self.loop = None
        self.session = None
        self.ready = None
        self.unmanaged_resources = []

    def _start(self):
        if self.session:
            return

        from ..remote.client import start_session
        try:
            self.session = start_session(self.url, extra={'env': self.env})
        except ConnectionRefusedError as e:
            raise ConnectionRefusedError(f"Could not connect to coordinator {self.url}") \
                from e

        self.loop = self.session.loop

    def on_resource_added(self, resource):
        if not isinstance(resource, RemotePlace):
            # we only need to handle new remote places here
            return
        remote_place = resource
        # Use the config file from the first resources we see (they should all
        # be the same).
        if not self.session:
            self.env = remote_place.target.env
            self.url = os.environ.get("LG_COORDINATOR", "127.0.0.1:20408")
            if self.env:
                config = self.env.config
                self.url = config.get_option("coordinator_address", self.url)
            self._start()
        place = self.session.get_place(remote_place.name)  # pylint: disable=no-member
        resource_entries = self.session.get_target_resources(place)  # pylint: disable=no-member
        expanded = []
        for (resource_name, _), resource_entry in resource_entries.items():
            new = target_factory.make_resource(
                remote_place.target, resource_entry.cls, resource_name, resource_entry.args)
            new.parent = remote_place
            new.avail = resource_entry.avail
            new.extra = resource_entry.extra
            new._remote_entry = resource_entry
            if not isinstance(new, ManagedResource):
                self.unmanaged_resources.append(new)
            expanded.append(new)
        self.logger.debug("expanded remote resources for place %s: %s", remote_place.name, expanded)
        remote_place.avail = True
        remote_place.tags = copy.deepcopy(place.tags)

    def poll(self):
        import asyncio
        if not self.loop.is_running():
            self.loop.run_until_complete(asyncio.sleep(0.1))
        for resource in self.resources + self.unmanaged_resources:
            if isinstance(resource, RemotePlace):
                continue
            attrs = resource._remote_entry.args.copy()
            attrs['avail'] = resource._remote_entry.avail
            # TODO allow the resource to do the update itself?
            changes = []
            fields = attr.fields(resource.__class__)
            for k, v_new in attrs.items():
                # check for attr converters
                attrib = getattr(fields, k)
                if attrib.converter:
                    v_new = attrib.converter(v_new)
                v_old = getattr(resource, k)
                setattr(resource, k, v_new)
                if v_old != v_new:
                    changes.append((k, v_old, v_new))
            if changes:
                self.logger.debug("changed attributes for %s:", resource)
                for k, v_old, v_new in changes:
                    self.logger.debug("  %s: %s -> %s", k, v_old, v_new)


@target_factory.reg_resource
@attr.s(eq=False)
class RemotePlace(ManagedResource):
    manager_cls = RemotePlaceManager

    def __attrs_post_init__(self):
        self.timeout = 10.0
        self.tags = {}
        super().__attrs_post_init__()

@attr.s(eq=False)
class RemoteUSBResource(NetworkResource, ManagedResource):
    manager_cls = RemotePlaceManager

    busnum = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(int)))
    devnum = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(int)))
    path = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(str)))
    vendor_id = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(int)))
    model_id = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(int)))


@target_factory.reg_resource
@attr.s(eq=False)
class RemoteAndroidUSBFastboot(RemoteUSBResource):
    def __attrs_post_init__(self):
        self.timeout = 10.0
        super().__attrs_post_init__()


@target_factory.reg_resource
@attr.s(eq=False)
class NetworkAndroidFastboot(RemoteAndroidUSBFastboot):
    def __attrs_post_init__(self):
        import warnings
        warnings.warn("NetworkAndroidFastboot is deprecated, use RemoteAndroidUSBFastboot instead",
                      DeprecationWarning)
        super().__attrs_post_init__()


@target_factory.reg_resource
@attr.s(eq=False)
class RemoteAndroidNetFastboot(NetworkResource):
    address = attr.ib(validator=attr.validators.instance_of(str))
    port = attr.ib(default=5554, validator=attr.validators.instance_of(int))
    protocol = attr.ib(default="udp", validator=attr.validators.instance_of(str))


@target_factory.reg_resource
@attr.s(eq=False)
class NetworkDFUDevice(RemoteUSBResource):
    def __attrs_post_init__(self):
        self.timeout = 10.0
        super().__attrs_post_init__()


@target_factory.reg_resource
@attr.s(eq=False)
class NetworkIMXUSBLoader(RemoteUSBResource):
    def __attrs_post_init__(self):
        self.timeout = 10.0
        super().__attrs_post_init__()


@target_factory.reg_resource
@attr.s(eq=False)
class NetworkMXSUSBLoader(RemoteUSBResource):
    def __attrs_post_init__(self):
        self.timeout = 10.0
        super().__attrs_post_init__()


@target_factory.reg_resource
@attr.s(eq=False)
class NetworkRKUSBLoader(RemoteUSBResource):
    def __attrs_post_init__(self):
        self.timeout = 10.0
        super().__attrs_post_init__()


@target_factory.reg_resource
@attr.s(eq=False)
class NetworkAlteraUSBBlaster(RemoteUSBResource):
    def __attrs_post_init__(self):
        self.timeout = 10.0
        super().__attrs_post_init__()


@target_factory.reg_resource
@attr.s(eq=False)
class NetworkSigrokUSBDevice(RemoteUSBResource):
    """The NetworkSigrokUSBDevice describes a remotely accessible sigrok USB device"""
    driver = attr.ib(
        default=None,
        validator=attr.validators.instance_of(str)
    )
    channels = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str))
    )
    channel_group = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str))
    )
    def __attrs_post_init__(self):
        self.timeout = 10.0
        super().__attrs_post_init__()


@target_factory.reg_resource
@attr.s(eq=False)
class NetworkSigrokUSBSerialDevice(RemoteUSBResource):
    """The NetworkSigrokUSBSerialDevice describes a remotely accessible sigrok USB device"""
    driver = attr.ib(
        default=None,
        validator=attr.validators.instance_of(str)
    )
    channels = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str))
    )
    channel_group = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str))
    )
    def __attrs_post_init__(self):
        self.timeout = 10.0
        super().__attrs_post_init__()


@target_factory.reg_resource
@attr.s(eq=False)
class NetworkUSBMassStorage(RemoteUSBResource):
    """The NetworkUSBMassStorage describes a remotely accessible USB storage device"""
    def __attrs_post_init__(self):
        self.timeout = 10.0
        super().__attrs_post_init__()


@target_factory.reg_resource
@attr.s(eq=False)
class NetworkUSBSDMuxDevice(RemoteUSBResource):
    """The NetworkUSBSDMuxDevice describes a remotely accessible USBSDMux device"""
    control_path = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str))
    )
    def __attrs_post_init__(self):
        self.timeout = 10.0
        super().__attrs_post_init__()


@target_factory.reg_resource
@attr.s(eq=False)
class NetworkUSBSDWireDevice(RemoteUSBResource):
    """The NetworkUSBSDWireDevice describes a remotely accessible USBSDWire device"""
    control_serial = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str))
    )
    def __attrs_post_init__(self):
        self.timeout = 10.0
        super().__attrs_post_init__()

@target_factory.reg_resource
@attr.s(eq=False)
class NetworkUSBSDWire3Device(RemoteUSBResource):
    """The NetworkUSBSDWire3Device describes a remotely accessible USBSDWire3 device"""
    control_serial = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str))
    )
    def __attrs_post_init__(self):
        self.timeout = 10.0
        super().__attrs_post_init__()

@target_factory.reg_resource
@attr.s(eq=False)
class NetworkSiSPMPowerPort(RemoteUSBResource):
    """The NetworkSiSPMPowerPort describes a remotely accessible SiS-PM power port"""
    index = attr.ib(default=None, validator=attr.validators.instance_of(int))
    def __attrs_post_init__(self):
        self.timeout = 10.0
        super().__attrs_post_init__()


@target_factory.reg_resource
@attr.s(eq=False)
class NetworkUSBPowerPort(RemoteUSBResource):
    """The NetworkUSBPowerPort describes a remotely accessible USB hub port with power switching"""
    index = attr.ib(default=None, validator=attr.validators.instance_of(int))
    def __attrs_post_init__(self):
        self.timeout = 10.0
        super().__attrs_post_init__()


@target_factory.reg_resource
@attr.s(eq=False)
class NetworkUSBVideo(RemoteUSBResource):
    """The NetworkUSBVideo describes a remotely accessible USB video device"""
    def __attrs_post_init__(self):
        self.timeout = 10.0
        super().__attrs_post_init__()


@target_factory.reg_resource
@attr.s(eq=False)
class NetworkUSBAudioInput(RemoteUSBResource):
    """The NetworkUSBAudioInput describes a remotely accessible USB audio input device"""
    index = attr.ib(default=0, validator=attr.validators.instance_of(int))
    alsa_name = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str))
    )
    def __attrs_post_init__(self):
        self.timeout = 10.0
        super().__attrs_post_init__()


@target_factory.reg_resource
@attr.s(eq=False)
class NetworkUSBTMC(RemoteUSBResource):
    """The NetworkUSBTMC describes a remotely accessible USB TMC device"""
    def __attrs_post_init__(self):
        self.timeout = 10.0
        super().__attrs_post_init__()


@target_factory.reg_resource
@attr.s(eq=False)
class NetworkUSBDebugger(RemoteUSBResource):
    """The NetworkUSBDebugger describes a remotely accessible USB JTAG/Debugger device"""
    def __attrs_post_init__(self):
        self.timeout = 10.0
        super().__attrs_post_init__()


@target_factory.reg_resource
@attr.s(eq=False)
class NetworkDeditecRelais8(RemoteUSBResource):
    """The NetworkDeditecRelais8 describes a remotely accessible USB relais port"""
    index = attr.ib(default=None, validator=attr.validators.instance_of(int))
    invert = attr.ib(default=False, validator=attr.validators.instance_of(bool))
    def __attrs_post_init__(self):
        self.timeout = 10.0
        super().__attrs_post_init__()


@target_factory.reg_resource
@attr.s(eq=False)
class NetworkHIDRelay(RemoteUSBResource):
    """The NetworkHIDRelay describes a remotely accessible USB relay port"""
    index = attr.ib(default=1, validator=attr.validators.instance_of(int))
    invert = attr.ib(default=False, validator=attr.validators.instance_of(bool))
    def __attrs_post_init__(self):
        self.timeout = 10.0
        super().__attrs_post_init__()


@target_factory.reg_resource
@attr.s(eq=False)
class NetworkSysfsGPIO(NetworkResource, ManagedResource):
    manager_cls = RemotePlaceManager

    """The NetworkSysfsGPIO describes a remotely accessible gpio line"""
    index = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(int)))
    def __attrs_post_init__(self):
        self.timeout = 10.0
        super().__attrs_post_init__()

@attr.s(eq=False)
class NetworkLXAIOBusNode(ManagedResource):
    manager_cls = RemotePlaceManager

    host = attr.ib(validator=attr.validators.instance_of(str))
    node = attr.ib(validator=attr.validators.instance_of(str))

    def __attrs_post_init__(self):
        self.timeout = 30.0
        super().__attrs_post_init__()

@target_factory.reg_resource
@attr.s(eq=False)
class NetworkLXAIOBusPIO(NetworkLXAIOBusNode):
    pin = attr.ib(validator=attr.validators.instance_of(str))
    invert = attr.ib(default=False, validator=attr.validators.instance_of(bool))

@target_factory.reg_resource
@attr.s(eq=False)
class NetworkLXAUSBMux(RemoteUSBResource):
    """The NetworkLXAUSBMux describes a remotely accessible USBMux device"""
    def __attrs_post_init__(self):
        self.timeout = 10.0
        super().__attrs_post_init__()

@target_factory.reg_resource
@attr.s(eq=False)
class NetworkUSBFlashableDevice(RemoteUSBResource):
    devnode = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(str)))

@attr.s(eq=False)
class NetworkMQTTResource(ManagedResource):
    manager_cls = RemotePlaceManager

    host = attr.ib(validator=attr.validators.instance_of(str))
    avail_topic = attr.ib(validator=attr.validators.instance_of(str))

    def __attrs_post_init__(self):
        self.timeout = 30.0
        super().__attrs_post_init__()

@target_factory.reg_resource
@attr.s(eq=False)
class RemoteNetworkInterface(NetworkResource, ManagedResource):
    manager_cls = RemotePlaceManager

    ifname = attr.ib(default=None)

@attr.s(eq=False)
class RemoteBaseProvider(NetworkResource):
    internal = attr.ib(validator=attr.validators.instance_of(str))
    external = attr.ib(validator=attr.validators.instance_of(str))


@target_factory.reg_resource
@attr.s(eq=False)
class RemoteTFTPProvider(RemoteBaseProvider):
    pass


@target_factory.reg_resource
@attr.s(eq=False)
class RemoteNFSProvider(NetworkResource):
    pass


@target_factory.reg_resource
@attr.s(eq=False)
class RemoteHTTPProvider(RemoteBaseProvider):
    pass



import attr

from ..factory import target_factory
from .common import Resource


@target_factory.reg_resource
@attr.s(eq=False)
class Eth008DigitalOutput(Resource):
    """This resource describes a digital output on an ETH008 relay board.

    Args:
        host (str): host to connect to
        index (str): index of the relay on the ETH008 board (1-8)
        invert (bool): whether to invert the output (default: False)
    """
    host = attr.ib(validator=attr.validators.instance_of(str))
    index = attr.ib(validator=attr.validators.instance_of(str),
                    converter=lambda x: str(int(x)))
    invert = attr.ib(default=False, validator=attr.validators.instance_of(bool))



import pytest
import socket

from labgrid import Environment
from labgrid.driver import SSHDriver, ExecutionError
from labgrid.exceptions import NoResourceFoundError
from labgrid.resource import NetworkService
from labgrid.util.helper import get_free_port

@pytest.fixture(scope='function')
def ssh_driver_mocked_and_activated(target, mocker):
    NetworkService(target, "service", "1.2.3.4", "root")
    call = mocker.patch('subprocess.call')
    call.return_value = 0
    popen = mocker.patch('subprocess.Popen', autospec=True)
    path = mocker.patch('os.path.exists')
    path.return_value = True
    instance_mock = mocker.MagicMock()
    popen.return_value = instance_mock
    instance_mock.wait = mocker.MagicMock(return_value=0)
    instance_mock.communicate = mocker.MagicMock(return_value=(b"", b""))
    SSHDriver(target, "ssh")
    s = target.get_driver("SSHDriver")
    return s

def test_create_fail_missing_resource(target):
    with pytest.raises(NoResourceFoundError):
        SSHDriver(target, "ssh")

def test_create(target, mocker):
    NetworkService(target, "service", "1.2.3.4", "root")
    call = mocker.patch('subprocess.call')
    call.return_value = 0
    popen = mocker.patch('subprocess.Popen', autospec=True)
    path = mocker.patch('os.path.exists')
    path.return_value = True
    instance_mock = mocker.MagicMock()
    popen.return_value = instance_mock
    instance_mock.wait = mocker.MagicMock(return_value=0)
    instance_mock.communicate = mocker.MagicMock(return_value=(b"", b""))
    s = SSHDriver(target, "ssh")
    assert isinstance(s, SSHDriver)

def test_run_check(target, ssh_driver_mocked_and_activated, mocker):
    s = ssh_driver_mocked_and_activated
    s._run = mocker.MagicMock(return_value=(['success'], [], 0))
    res = s.run_check("test")
    assert res == ['success']
    res = s.run("test")
    assert res == (['success'], [], 0)
    target.deactivate(s)

def test_run_check_raise(target, ssh_driver_mocked_and_activated, mocker):
    s = ssh_driver_mocked_and_activated
    s._run = mocker.MagicMock(return_value=(['error'], [], 1))
    with pytest.raises(ExecutionError):
        res = s.run_check("test")
    res = s.run("test")
    assert res == (['error'], [], 1)
    target.deactivate(s)

def test_default_tools(target):
    NetworkService(target, "service", "1.2.3.4", "root")
    s = SSHDriver(target, "ssh")
    assert [s._ssh, s._scp, s._sshfs, s._rsync] == ["ssh", "scp", "sshfs", "rsync"]

def test_custom_tools(target, tmpdir):
    p = tmpdir.join("config.yaml")
    p.write(
        """
        tools:
          ssh: "/path/to/ssh"
          scp: "/path/to/scp"
          sshfs: "/path/to/sshfs"
          rsync: "/path/to/rsync"
        """
    )
    target.env = Environment(str(p))
    NetworkService(target, "service", "1.2.3.4", "root")
    s = SSHDriver(target, "ssh")
    assert [s._ssh, s._scp, s._sshfs, s._rsync] == [f"/path/to/{t}" for t in ("ssh", "scp", "sshfs", "rsync")]

@pytest.fixture(scope='function')
def ssh_localhost(target, pytestconfig):
    name = pytestconfig.getoption("--ssh-username")
    NetworkService(target, "service", "localhost", name)
    SSHDriver(target, "ssh")
    s = target.get_driver("SSHDriver")
    return s

@pytest.mark.sshusername
def test_local_put(ssh_localhost, tmpdir):
    p = tmpdir.join("config.yaml")
    p.write(
        """PUT Teststring"""
    )

    ssh_localhost.put(str(p), "/tmp/test_put_yaml")
    assert open('/tmp/test_put_yaml', 'r').readlines() == [ "PUT Teststring" ]

@pytest.mark.sshusername
def test_local_no_final_line_remove(ssh_localhost, tmpdir):
    p = tmpdir.join("test_no_line_remove")
    p.write("teststring")

    ssh_localhost.put(str(p), "/tmp/test_line_remove")
    stdout, stderr, _ = ssh_localhost.run("cat /tmp/test_line_remove")
    assert stdout == [ "teststring" ]

@pytest.mark.sshusername
def test_local_final_line_remove_empty(ssh_localhost, tmpdir):
    p = tmpdir.join("test_no_line_remove")
    p.write("teststring\n")

    ssh_localhost.put(str(p), "/tmp/test_line_remove")
    stdout, stderr, _ = ssh_localhost.run("cat /tmp/test_line_remove")
    assert stdout == [ "teststring" ]

@pytest.mark.sshusername
def test_local_put_dir(ssh_localhost, tmpdir):
    d = tmpdir.mkdir("test_put_dir");
    p = d.join("config.yaml")
    p.write(
        """PUT Teststring"""
    )

    ssh_localhost.put(str(d), "/tmp/test_put_dir")
    assert open('/tmp/test_put_dir/config.yaml', 'r').readlines() == [ "PUT Teststring" ]

@pytest.mark.sshusername
def test_local_get(ssh_localhost, tmpdir):
    p = tmpdir.join("config.yaml")
    p.write(
        """GET Teststring"""
    )

    ssh_localhost.get(str(p), "/tmp/test_get_yaml")
    assert open('/tmp/test_get_yaml', 'r').readlines() == [ "GET Teststring" ]

@pytest.mark.sshusername
def test_local_get_dir(ssh_localhost, tmpdir):
    d = tmpdir.mkdir("test_get_dir");
    p = d.join("config.yaml")
    p.write(
        """GET Teststring"""
    )

    ssh_localhost.get(str(d), "/tmp/test_get_dir")
    assert open('/tmp/test_get_dir/config.yaml', 'r').readlines() == [ "GET Teststring" ]

@pytest.mark.sshusername
def test_local_run(ssh_localhost, tmpdir):

    res = ssh_localhost.run("echo Hello")
    assert res == (["Hello"], [], 0)

@pytest.mark.sshusername
def test_local_run_check(ssh_localhost, tmpdir):

    res = ssh_localhost.run_check("echo Hello")
    assert res == (["Hello"])

@pytest.mark.sshusername
def test_local_port_forward(ssh_localhost, tmpdir):
    remoteport = get_free_port()
    test_string = "Hello World"

    with ssh_localhost.forward_local_port(remoteport) as localport:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as server_socket:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as send_socket:
                server_socket.bind(("127.0.0.1", remoteport))
                server_socket.listen(1)

                send_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                send_socket.connect(("127.0.0.1", localport))

                client_socket, address = server_socket.accept()
                send_socket.send(test_string.encode('utf-8'))

                assert client_socket.recv(16).decode("utf-8") == test_string

@pytest.mark.sshusername
def test_local_remote_forward(ssh_localhost, tmpdir):
    remoteport = get_free_port()
    localport = get_free_port()
    test_string = "Hello World"

    with ssh_localhost.forward_remote_port(remoteport, localport):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as server_socket:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as send_socket:
                server_socket.bind(("127.0.0.1", localport))
                server_socket.listen(1)

                send_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                send_socket.connect(("127.0.0.1", remoteport))

                client_socket, address = server_socket.accept()
                send_socket.send(test_string.encode('utf-8'))

                assert client_socket.recv(16).decode("utf-8") == test_string


@pytest.mark.sshusername
def test_unix_socket_forward(ssh_localhost, tmpdir):
    p = tmpdir.join("console.sock")
    test_string = "Hello World"

    with ssh_localhost.forward_unix_socket(str(p)) as localport:
        with socket.socket(socket.AF_UNIX, socket.SOCK_STREAM) as server_socket:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as send_socket:
                server_socket.bind(str(p))
                server_socket.listen(1)

                send_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                send_socket.connect(("127.0.0.1", localport))

                client_socket, address = server_socket.accept()
                send_socket.send(test_string.encode("utf-8"))

                assert client_socket.recv(16).decode("utf-8") == test_string



import attr

from ..factory import target_factory
from .common import Resource


@attr.s(eq=False)
class SerialPort(Resource):
    """The basic SerialPort describes port and speed

    Args:
        port (str): port to connect to
        speed (int): speed of the port, defaults to 115200"""
    port = attr.ib(default=None)
    speed = attr.ib(default=115200, validator=attr.validators.instance_of(int))


@target_factory.reg_resource
@attr.s(eq=False)
class NetworkInterface(Resource):
    """The basic NetworkInterface contains an interface name

    Args:
        ifname (str): name of the interface"""
    ifname = attr.ib(default=None)

@target_factory.reg_resource
@attr.s
class EthernetPort(Resource):
    """The basic EthernetPort describes a switch and interface

    Args:
        switch (str): name of the switch
        interface (str): name of the interface"""
    switch = attr.ib(default=None)
    interface = attr.ib(default=None)


@target_factory.reg_resource
@attr.s(eq=False)
class SysfsGPIO(Resource):
    """The basic SysfsGPIO contains an index

    Args:
        index (int): index of target gpio line."""
    index = attr.ib(default=None, validator=attr.validators.instance_of(int))



import pytest
import stat
import os
from labgrid.consoleloggingreporter import ConsoleLoggingReporter


@pytest.fixture(scope="function")
def consolelogger(tmpdir):
    ConsoleLoggingReporter.start(str(tmpdir))
    yield
    ConsoleLoggingReporter.stop()


def test_consoleloggingreporter_output_with_name(consolelogger, serial_driver, tmpdir):
    def return_test(self, size=1, timeout=0.0):
        return b"test"

    serial_driver.serial.in_waiting = 4
    serial_driver.serial.read = return_test
    serial_driver.read()
    assert tmpdir.join("console_Test_serial").readlines()[-1] == "test"


def test_consoleloggingreporter_output_without_name(consolelogger, serial_driver_no_name, tmpdir):
    def return_test(self, size=1, timeout=0.0):
        return b"test"

    serial_driver_no_name.serial.in_waiting = 4
    serial_driver_no_name.serial.read = return_test
    serial_driver_no_name.read()
    assert tmpdir.join("console_Test").readlines()[-1] == "test"


def test_consoleloggingreporter_dir_not_writeable(consolelogger, serial_driver, tmpdir):
    os.chmod(str(tmpdir), stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH)

    def return_test(self, size=1, timeout=0.0):
        return b"test"

    serial_driver.serial.in_waiting = 4
    serial_driver.serial.read = return_test
    serial_driver.read()



#!/usr/bin/env python3

import time

import attr

from .common import Driver
from ..factory import target_factory
from ..protocol import PowerProtocol
from ..step import step
from ..util import Timeout


class MQTTError(Exception):
    pass


@target_factory.reg_driver
@attr.s(eq=False)
class TasmotaPowerDriver(Driver, PowerProtocol):
    bindings = {"power": {"TasmotaPowerPort"}}
    delay = attr.ib(default=2.0, validator=attr.validators.instance_of(float))
    _client = attr.ib(default=None)
    _status = attr.ib(default=None)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        import paho.mqtt.client as mqtt

        self._client = mqtt.Client(mqtt.CallbackAPIVersion.VERSION2)

    def on_activate(self):
        self._client.on_message = self._on_message
        self._client.on_connect = self._on_connect
        self._client.connect(self.power.host)
        self._client.loop_start()

    def on_deactivate(self):
        self._client.loop_stop()

    def _on_message(self, client, userdata, msg):
        if msg.payload == b"ON":
            status = True
        elif msg.payload == b"OFF":
            status = False
        else:
            raise ValueError(f"Unknown status: {msg.payload}. Must be 'ON' or 'OFF'")
        self._status = status

    def _on_connect(self, client, userdata, flags, reason_code, properties):
        client.subscribe(self.power.status_topic)

    def _publish(self, topic, payload):
        msg = self._client.publish(topic, payload=payload)
        timeout = Timeout(3.0)
        while not msg.is_published:
            time.sleep(0.1)
            if timeout.expired:
                raise MQTTError("publish timed out")
        return msg

    @Driver.check_active
    @step()
    def on(self):
        self._publish(self.power.power_topic, "ON")
        timeout = Timeout(3.0)
        while self._status is False:
            time.sleep(0.1)
            if timeout.expired:
                raise MQTTError("Port did not change status within 3 seconds")

    @Driver.check_active
    @step()
    def off(self):
        self._publish(self.power.power_topic, "OFF")
        timeout = Timeout(3.0)
        while self._status is True:
            time.sleep(0.1)
            if timeout.expired:
                raise MQTTError("Port did not change status within 3 seconds")

    @Driver.check_active
    @step()
    def cycle(self):
        self.off()
        time.sleep(self.delay)
        self.on()

    @Driver.check_active
    @step()
    def get(self):
        self._client.publish(self.power.power_topic)
        timeout = Timeout(3.0)
        while self._status is None:
            time.sleep(0.1)
            if timeout.expired:
                raise MQTTError("Could not get initial status")
        return self._status



import pytest
from shutil import which
import subprocess

from labgrid.resource.udev import USBDebugger
from labgrid.driver.openocddriver import OpenOCDDriver


pytestmark = pytest.mark.skipif(not which("openocd"), reason="openocd not available")


def test_openocd_resource(target):
    r = USBDebugger(target, name=None, match={"sys_name": "1-12"})


def test_openocd_driver_activate(target):
    r = USBDebugger(target, name=None, match={"sys_name": "1-12"})
    r.avail = True
    d = OpenOCDDriver(target, name=None)
    target.activate(d)


def test_openocd_driver(target, tmpdir):
    r = USBDebugger(target, name=None, match={"sys_name": "1-12"})
    r.avail = True
    d = OpenOCDDriver(target, name=None, load_commands=["shutdown"])
    target.activate(d)
    d.load(__file__)

    with pytest.raises(subprocess.CalledProcessError):
        d.execute(["invalid_command_labgrid"])

    d.execute(["shutdown"])



from labgrid import Target
from labgrid.resource import USBPowerPort
from labgrid.driver import USBPowerDriver


t = Target(name="main")
upp = USBPowerPort(t, name=None, match={"ID_PATH": "pci-0000:00:14.0-usb-0:2:1.0"}, index=1)
upd = USBPowerDriver(t, name=None)

t.activate(upd)

print("Target:", t)
print("  Resources:", t.resources)
print("  Drivers:", t.drivers)
print("Status:", upd.get())

print("\n=== Switching power OFF ===")
upd.off()
print("Status:", upd.get())

print("\n=== Switching power ON ===")
upd.on()
print("Status:", upd.get())



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class ClientInMessage(_message.Message):
    __slots__ = ("sync", "startup", "subscribe")
    SYNC_FIELD_NUMBER: _ClassVar[int]
    STARTUP_FIELD_NUMBER: _ClassVar[int]
    SUBSCRIBE_FIELD_NUMBER: _ClassVar[int]
    sync: Sync
    startup: StartupDone
    subscribe: Subscribe
    def __init__(self, sync: _Optional[_Union[Sync, _Mapping]] = ..., startup: _Optional[_Union[StartupDone, _Mapping]] = ..., subscribe: _Optional[_Union[Subscribe, _Mapping]] = ...) -> None: ...

class Sync(_message.Message):
    __slots__ = ("id",)
    ID_FIELD_NUMBER: _ClassVar[int]
    id: int
    def __init__(self, id: _Optional[int] = ...) -> None: ...

class StartupDone(_message.Message):
    __slots__ = ("version", "name")
    VERSION_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    version: str
    name: str
    def __init__(self, version: _Optional[str] = ..., name: _Optional[str] = ...) -> None: ...

class Subscribe(_message.Message):
    __slots__ = ("is_unsubscribe", "all_places", "all_resources")
    IS_UNSUBSCRIBE_FIELD_NUMBER: _ClassVar[int]
    ALL_PLACES_FIELD_NUMBER: _ClassVar[int]
    ALL_RESOURCES_FIELD_NUMBER: _ClassVar[int]
    is_unsubscribe: bool
    all_places: bool
    all_resources: bool
    def __init__(self, is_unsubscribe: bool = ..., all_places: bool = ..., all_resources: bool = ...) -> None: ...

class ClientOutMessage(_message.Message):
    __slots__ = ("sync", "updates")
    SYNC_FIELD_NUMBER: _ClassVar[int]
    UPDATES_FIELD_NUMBER: _ClassVar[int]
    sync: Sync
    updates: _containers.RepeatedCompositeFieldContainer[UpdateResponse]
    def __init__(self, sync: _Optional[_Union[Sync, _Mapping]] = ..., updates: _Optional[_Iterable[_Union[UpdateResponse, _Mapping]]] = ...) -> None: ...

class UpdateResponse(_message.Message):
    __slots__ = ("resource", "del_resource", "place", "del_place")
    RESOURCE_FIELD_NUMBER: _ClassVar[int]
    DEL_RESOURCE_FIELD_NUMBER: _ClassVar[int]
    PLACE_FIELD_NUMBER: _ClassVar[int]
    DEL_PLACE_FIELD_NUMBER: _ClassVar[int]
    resource: Resource
    del_resource: Resource.Path
    place: Place
    del_place: str
    def __init__(self, resource: _Optional[_Union[Resource, _Mapping]] = ..., del_resource: _Optional[_Union[Resource.Path, _Mapping]] = ..., place: _Optional[_Union[Place, _Mapping]] = ..., del_place: _Optional[str] = ...) -> None: ...

class ExporterInMessage(_message.Message):
    __slots__ = ("resource", "startup", "response")
    RESOURCE_FIELD_NUMBER: _ClassVar[int]
    STARTUP_FIELD_NUMBER: _ClassVar[int]
    RESPONSE_FIELD_NUMBER: _ClassVar[int]
    resource: Resource
    startup: StartupDone
    response: ExporterResponse
    def __init__(self, resource: _Optional[_Union[Resource, _Mapping]] = ..., startup: _Optional[_Union[StartupDone, _Mapping]] = ..., response: _Optional[_Union[ExporterResponse, _Mapping]] = ...) -> None: ...

class Resource(_message.Message):
    __slots__ = ("path", "cls", "params", "extra", "acquired", "avail")
    class Path(_message.Message):
        __slots__ = ("exporter_name", "group_name", "resource_name")
        EXPORTER_NAME_FIELD_NUMBER: _ClassVar[int]
        GROUP_NAME_FIELD_NUMBER: _ClassVar[int]
        RESOURCE_NAME_FIELD_NUMBER: _ClassVar[int]
        exporter_name: str
        group_name: str
        resource_name: str
        def __init__(self, exporter_name: _Optional[str] = ..., group_name: _Optional[str] = ..., resource_name: _Optional[str] = ...) -> None: ...
    class ParamsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: MapValue
        def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[MapValue, _Mapping]] = ...) -> None: ...
    class ExtraEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: MapValue
        def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[MapValue, _Mapping]] = ...) -> None: ...
    PATH_FIELD_NUMBER: _ClassVar[int]
    CLS_FIELD_NUMBER: _ClassVar[int]
    PARAMS_FIELD_NUMBER: _ClassVar[int]
    EXTRA_FIELD_NUMBER: _ClassVar[int]
    ACQUIRED_FIELD_NUMBER: _ClassVar[int]
    AVAIL_FIELD_NUMBER: _ClassVar[int]
    path: Resource.Path
    cls: str
    params: _containers.MessageMap[str, MapValue]
    extra: _containers.MessageMap[str, MapValue]
    acquired: str
    avail: bool
    def __init__(self, path: _Optional[_Union[Resource.Path, _Mapping]] = ..., cls: _Optional[str] = ..., params: _Optional[_Mapping[str, MapValue]] = ..., extra: _Optional[_Mapping[str, MapValue]] = ..., acquired: _Optional[str] = ..., avail: bool = ...) -> None: ...

class MapValue(_message.Message):
    __slots__ = ("bool_value", "int_value", "uint_value", "float_value", "string_value")
    BOOL_VALUE_FIELD_NUMBER: _ClassVar[int]
    INT_VALUE_FIELD_NUMBER: _ClassVar[int]
    UINT_VALUE_FIELD_NUMBER: _ClassVar[int]
    FLOAT_VALUE_FIELD_NUMBER: _ClassVar[int]
    STRING_VALUE_FIELD_NUMBER: _ClassVar[int]
    bool_value: bool
    int_value: int
    uint_value: int
    float_value: float
    string_value: str
    def __init__(self, bool_value: bool = ..., int_value: _Optional[int] = ..., uint_value: _Optional[int] = ..., float_value: _Optional[float] = ..., string_value: _Optional[str] = ...) -> None: ...

class ExporterResponse(_message.Message):
    __slots__ = ("success", "reason")
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    REASON_FIELD_NUMBER: _ClassVar[int]
    success: bool
    reason: str
    def __init__(self, success: bool = ..., reason: _Optional[str] = ...) -> None: ...

class Hello(_message.Message):
    __slots__ = ("version",)
    VERSION_FIELD_NUMBER: _ClassVar[int]
    version: str
    def __init__(self, version: _Optional[str] = ...) -> None: ...

class ExporterOutMessage(_message.Message):
    __slots__ = ("hello", "set_acquired_request")
    HELLO_FIELD_NUMBER: _ClassVar[int]
    SET_ACQUIRED_REQUEST_FIELD_NUMBER: _ClassVar[int]
    hello: Hello
    set_acquired_request: ExporterSetAcquiredRequest
    def __init__(self, hello: _Optional[_Union[Hello, _Mapping]] = ..., set_acquired_request: _Optional[_Union[ExporterSetAcquiredRequest, _Mapping]] = ...) -> None: ...

class ExporterSetAcquiredRequest(_message.Message):
    __slots__ = ("group_name", "resource_name", "place_name")
    GROUP_NAME_FIELD_NUMBER: _ClassVar[int]
    RESOURCE_NAME_FIELD_NUMBER: _ClassVar[int]
    PLACE_NAME_FIELD_NUMBER: _ClassVar[int]
    group_name: str
    resource_name: str
    place_name: str
    def __init__(self, group_name: _Optional[str] = ..., resource_name: _Optional[str] = ..., place_name: _Optional[str] = ...) -> None: ...

class AddPlaceRequest(_message.Message):
    __slots__ = ("name",)
    NAME_FIELD_NUMBER: _ClassVar[int]
    name: str
    def __init__(self, name: _Optional[str] = ...) -> None: ...

class AddPlaceResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class DeletePlaceRequest(_message.Message):
    __slots__ = ("name",)
    NAME_FIELD_NUMBER: _ClassVar[int]
    name: str
    def __init__(self, name: _Optional[str] = ...) -> None: ...

class DeletePlaceResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetPlacesRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetPlacesResponse(_message.Message):
    __slots__ = ("places",)
    PLACES_FIELD_NUMBER: _ClassVar[int]
    places: _containers.RepeatedCompositeFieldContainer[Place]
    def __init__(self, places: _Optional[_Iterable[_Union[Place, _Mapping]]] = ...) -> None: ...

class Place(_message.Message):
    __slots__ = ("name", "aliases", "comment", "tags", "matches", "acquired", "acquired_resources", "allowed", "created", "changed", "reservation")
    class TagsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    NAME_FIELD_NUMBER: _ClassVar[int]
    ALIASES_FIELD_NUMBER: _ClassVar[int]
    COMMENT_FIELD_NUMBER: _ClassVar[int]
    TAGS_FIELD_NUMBER: _ClassVar[int]
    MATCHES_FIELD_NUMBER: _ClassVar[int]
    ACQUIRED_FIELD_NUMBER: _ClassVar[int]
    ACQUIRED_RESOURCES_FIELD_NUMBER: _ClassVar[int]
    ALLOWED_FIELD_NUMBER: _ClassVar[int]
    CREATED_FIELD_NUMBER: _ClassVar[int]
    CHANGED_FIELD_NUMBER: _ClassVar[int]
    RESERVATION_FIELD_NUMBER: _ClassVar[int]
    name: str
    aliases: _containers.RepeatedScalarFieldContainer[str]
    comment: str
    tags: _containers.ScalarMap[str, str]
    matches: _containers.RepeatedCompositeFieldContainer[ResourceMatch]
    acquired: str
    acquired_resources: _containers.RepeatedScalarFieldContainer[str]
    allowed: _containers.RepeatedScalarFieldContainer[str]
    created: float
    changed: float
    reservation: str
    def __init__(self, name: _Optional[str] = ..., aliases: _Optional[_Iterable[str]] = ..., comment: _Optional[str] = ..., tags: _Optional[_Mapping[str, str]] = ..., matches: _Optional[_Iterable[_Union[ResourceMatch, _Mapping]]] = ..., acquired: _Optional[str] = ..., acquired_resources: _Optional[_Iterable[str]] = ..., allowed: _Optional[_Iterable[str]] = ..., created: _Optional[float] = ..., changed: _Optional[float] = ..., reservation: _Optional[str] = ...) -> None: ...

class ResourceMatch(_message.Message):
    __slots__ = ("exporter", "group", "cls", "name", "rename")
    EXPORTER_FIELD_NUMBER: _ClassVar[int]
    GROUP_FIELD_NUMBER: _ClassVar[int]
    CLS_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    RENAME_FIELD_NUMBER: _ClassVar[int]
    exporter: str
    group: str
    cls: str
    name: str
    rename: str
    def __init__(self, exporter: _Optional[str] = ..., group: _Optional[str] = ..., cls: _Optional[str] = ..., name: _Optional[str] = ..., rename: _Optional[str] = ...) -> None: ...

class AddPlaceAliasRequest(_message.Message):
    __slots__ = ("placename", "alias")
    PLACENAME_FIELD_NUMBER: _ClassVar[int]
    ALIAS_FIELD_NUMBER: _ClassVar[int]
    placename: str
    alias: str
    def __init__(self, placename: _Optional[str] = ..., alias: _Optional[str] = ...) -> None: ...

class AddPlaceAliasResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class DeletePlaceAliasRequest(_message.Message):
    __slots__ = ("placename", "alias")
    PLACENAME_FIELD_NUMBER: _ClassVar[int]
    ALIAS_FIELD_NUMBER: _ClassVar[int]
    placename: str
    alias: str
    def __init__(self, placename: _Optional[str] = ..., alias: _Optional[str] = ...) -> None: ...

class DeletePlaceAliasResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class SetPlaceTagsRequest(_message.Message):
    __slots__ = ("placename", "tags")
    class TagsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    PLACENAME_FIELD_NUMBER: _ClassVar[int]
    TAGS_FIELD_NUMBER: _ClassVar[int]
    placename: str
    tags: _containers.ScalarMap[str, str]
    def __init__(self, placename: _Optional[str] = ..., tags: _Optional[_Mapping[str, str]] = ...) -> None: ...

class SetPlaceTagsResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class SetPlaceCommentRequest(_message.Message):
    __slots__ = ("placename", "comment")
    PLACENAME_FIELD_NUMBER: _ClassVar[int]
    COMMENT_FIELD_NUMBER: _ClassVar[int]
    placename: str
    comment: str
    def __init__(self, placename: _Optional[str] = ..., comment: _Optional[str] = ...) -> None: ...

class SetPlaceCommentResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class AddPlaceMatchRequest(_message.Message):
    __slots__ = ("placename", "pattern", "rename")
    PLACENAME_FIELD_NUMBER: _ClassVar[int]
    PATTERN_FIELD_NUMBER: _ClassVar[int]
    RENAME_FIELD_NUMBER: _ClassVar[int]
    placename: str
    pattern: str
    rename: str
    def __init__(self, placename: _Optional[str] = ..., pattern: _Optional[str] = ..., rename: _Optional[str] = ...) -> None: ...

class AddPlaceMatchResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class DeletePlaceMatchRequest(_message.Message):
    __slots__ = ("placename", "pattern", "rename")
    PLACENAME_FIELD_NUMBER: _ClassVar[int]
    PATTERN_FIELD_NUMBER: _ClassVar[int]
    RENAME_FIELD_NUMBER: _ClassVar[int]
    placename: str
    pattern: str
    rename: str
    def __init__(self, placename: _Optional[str] = ..., pattern: _Optional[str] = ..., rename: _Optional[str] = ...) -> None: ...

class DeletePlaceMatchResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class AcquirePlaceRequest(_message.Message):
    __slots__ = ("placename",)
    PLACENAME_FIELD_NUMBER: _ClassVar[int]
    placename: str
    def __init__(self, placename: _Optional[str] = ...) -> None: ...

class AcquirePlaceResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class ReleasePlaceRequest(_message.Message):
    __slots__ = ("placename", "fromuser")
    PLACENAME_FIELD_NUMBER: _ClassVar[int]
    FROMUSER_FIELD_NUMBER: _ClassVar[int]
    placename: str
    fromuser: str
    def __init__(self, placename: _Optional[str] = ..., fromuser: _Optional[str] = ...) -> None: ...

class ReleasePlaceResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class AllowPlaceRequest(_message.Message):
    __slots__ = ("placename", "user")
    PLACENAME_FIELD_NUMBER: _ClassVar[int]
    USER_FIELD_NUMBER: _ClassVar[int]
    placename: str
    user: str
    def __init__(self, placename: _Optional[str] = ..., user: _Optional[str] = ...) -> None: ...

class AllowPlaceResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class CreateReservationRequest(_message.Message):
    __slots__ = ("filters", "prio")
    class FiltersEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: Reservation.Filter
        def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[Reservation.Filter, _Mapping]] = ...) -> None: ...
    FILTERS_FIELD_NUMBER: _ClassVar[int]
    PRIO_FIELD_NUMBER: _ClassVar[int]
    filters: _containers.MessageMap[str, Reservation.Filter]
    prio: float
    def __init__(self, filters: _Optional[_Mapping[str, Reservation.Filter]] = ..., prio: _Optional[float] = ...) -> None: ...

class CreateReservationResponse(_message.Message):
    __slots__ = ("reservation",)
    RESERVATION_FIELD_NUMBER: _ClassVar[int]
    reservation: Reservation
    def __init__(self, reservation: _Optional[_Union[Reservation, _Mapping]] = ...) -> None: ...

class Reservation(_message.Message):
    __slots__ = ("owner", "token", "state", "prio", "filters", "allocations", "created", "timeout")
    class Filter(_message.Message):
        __slots__ = ("filter",)
        class FilterEntry(_message.Message):
            __slots__ = ("key", "value")
            KEY_FIELD_NUMBER: _ClassVar[int]
            VALUE_FIELD_NUMBER: _ClassVar[int]
            key: str
            value: str
            def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
        FILTER_FIELD_NUMBER: _ClassVar[int]
        filter: _containers.ScalarMap[str, str]
        def __init__(self, filter: _Optional[_Mapping[str, str]] = ...) -> None: ...
    class FiltersEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: Reservation.Filter
        def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[Reservation.Filter, _Mapping]] = ...) -> None: ...
    class AllocationsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    OWNER_FIELD_NUMBER: _ClassVar[int]
    TOKEN_FIELD_NUMBER: _ClassVar[int]
    STATE_FIELD_NUMBER: _ClassVar[int]
    PRIO_FIELD_NUMBER: _ClassVar[int]
    FILTERS_FIELD_NUMBER: _ClassVar[int]
    ALLOCATIONS_FIELD_NUMBER: _ClassVar[int]
    CREATED_FIELD_NUMBER: _ClassVar[int]
    TIMEOUT_FIELD_NUMBER: _ClassVar[int]
    owner: str
    token: str
    state: int
    prio: float
    filters: _containers.MessageMap[str, Reservation.Filter]
    allocations: _containers.ScalarMap[str, str]
    created: float
    timeout: float
    def __init__(self, owner: _Optional[str] = ..., token: _Optional[str] = ..., state: _Optional[int] = ..., prio: _Optional[float] = ..., filters: _Optional[_Mapping[str, Reservation.Filter]] = ..., allocations: _Optional[_Mapping[str, str]] = ..., created: _Optional[float] = ..., timeout: _Optional[float] = ...) -> None: ...

class CancelReservationRequest(_message.Message):
    __slots__ = ("token",)
    TOKEN_FIELD_NUMBER: _ClassVar[int]
    token: str
    def __init__(self, token: _Optional[str] = ...) -> None: ...

class CancelReservationResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class PollReservationRequest(_message.Message):
    __slots__ = ("token",)
    TOKEN_FIELD_NUMBER: _ClassVar[int]
    token: str
    def __init__(self, token: _Optional[str] = ...) -> None: ...

class PollReservationResponse(_message.Message):
    __slots__ = ("reservation",)
    RESERVATION_FIELD_NUMBER: _ClassVar[int]
    reservation: Reservation
    def __init__(self, reservation: _Optional[_Union[Reservation, _Mapping]] = ...) -> None: ...

class GetReservationsResponse(_message.Message):
    __slots__ = ("reservations",)
    RESERVATIONS_FIELD_NUMBER: _ClassVar[int]
    reservations: _containers.RepeatedCompositeFieldContainer[Reservation]
    def __init__(self, reservations: _Optional[_Iterable[_Union[Reservation, _Mapping]]] = ...) -> None: ...

class GetReservationsRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...



def test_ssh(target):
    """Simple test with SSHDriver directly executing a command on the DUT"""
    ssh_driver = target.get_driver("SSHDriver")
    stdout = ssh_driver.run_check("cat /proc/version")
    assert stdout



"""
This module implements switching GPIOs via sysfs GPIO kernel interface.

Takes an integer property 'index' which refers to the already exported GPIO device.

"""
import logging
import os

class GpioDigitalOutput:
    _gpio_sysfs_path_prefix = '/sys/class/gpio'

    @staticmethod
    def _assert_gpio_line_is_exported(index):
        gpio_sysfs_path = os.path.join(GpioDigitalOutput._gpio_sysfs_path_prefix,
                                       f'gpio{index}')
        # Deprecated: the exporter can export on acquire, we are leaving this
        # in for now to support exporters which have not been updated yet.
        if not os.path.exists(gpio_sysfs_path):
            export_sysfs_path = os.path.join(GpioDigitalOutput._gpio_sysfs_path_prefix, 'export')
            with open(export_sysfs_path, mode='wb') as export:
                export.write(str(index).encode('utf-8'))
        if not os.path.exists(gpio_sysfs_path):
            raise ValueError("Device not found")

    def __init__(self, index):
        self.gpio_sysfs_value_fd = None
        self._logger = logging.getLogger("Device: ")
        GpioDigitalOutput._assert_gpio_line_is_exported(index)
        gpio_sysfs_path = os.path.join(GpioDigitalOutput._gpio_sysfs_path_prefix,
                                       f'gpio{index}')

        gpio_sysfs_direction_path = os.path.join(gpio_sysfs_path, 'direction')
        with open(gpio_sysfs_direction_path, 'rb') as direction_fd:
            literal_value = direction_fd.read(3)
        if literal_value != b"out":
            self._logger.debug("Configuring GPIO %d as output.", index)
            with open(gpio_sysfs_direction_path, 'wb') as direction_fd:
                direction_fd.write(b'out')

        gpio_sysfs_value_path = os.path.join(gpio_sysfs_path, 'value')
        self.gpio_sysfs_value_fd = os.open(gpio_sysfs_value_path, flags=(os.O_RDWR | os.O_SYNC))

    def __del__(self):
        if self.gpio_sysfs_value_fd:
            os.close(self.gpio_sysfs_value_fd)
        self.gpio_sysfs_value_fd = None

    def get(self):
        os.lseek(self.gpio_sysfs_value_fd, 0, os.SEEK_SET)
        literal_value = os.read(self.gpio_sysfs_value_fd, 1)
        if literal_value == b'0':
            return False
        elif literal_value == b'1':
            return True
        raise ValueError("GPIO value is out of range.")

    def set(self, status):
        self._logger.debug("Setting GPIO to `%s`.", status)
        binary_value = None
        if status is True:
            binary_value = b'1'
        elif status is False:
            binary_value = b'0'

        if binary_value is None:
            raise ValueError("GPIO value is out of range.")

        os.write(self.gpio_sysfs_value_fd, binary_value)


_gpios = {}

def _get_gpio_line(index):
    if index not in _gpios:
        _gpios[index] = GpioDigitalOutput(index=index)
    return _gpios[index]

def handle_set(index, status):
    gpio_line = _get_gpio_line(index)
    gpio_line.set(status)


def handle_get(index):
    gpio_line = _get_gpio_line(index)
    return gpio_line.get()


methods = {
    'set': handle_set,
    'get': handle_get,
}



from .common import Strategy, StrategyError, never_retry
from .bareboxstrategy import BareboxStrategy
from .shellstrategy import ShellStrategy
from .ubootstrategy import UBootStrategy
from .graphstrategy import *
from .dockerstrategy import DockerStrategy



"""
    Controls the *Ubiquity mFi mPower* Power Strip with Ethernet and Wi-Fi connectivity via HTTP.
    Reference: https://dl.ubnt.com/guides/mfi/mFi_mPower_PRO_US_QSG.pdf

    Example configuration to use port #3 on a device with URL 'http://172.17.180.53/'
    with the default credentials ('ubnt' for both username and password):

    NetworkPowerPort:
      model: mfi_mpower
      host: 'http://172.17.180.53/'
      index: 3

    Custom credentials can be provided in the URL itself:

    NetworkPowerPort:
      model: mfi_mpower
      host: 'http://username:password@172.17.180.53/'
      index: 3
"""

from typing import Tuple
from urllib.parse import urlparse, urljoin

import requests

from ..exception import ExecutionError


def login(s: requests.Session, base_url: str, credentials: dict) -> None:

    # We need to first fetch the base url to satisfy the Cookie Monster
    s.get(base_url)

    s.post(urljoin(base_url, '/login.cgi'),
        data=dict(username=credentials['username'], password=credentials['password']))


# Obtain credentials and repack base_url if needed
def get_credentials(base_url: str) -> Tuple[str, dict]:
    base_url = urlparse(base_url)

    if base_url.username is None or base_url.password is None:
        credentials = dict(username='ubnt', password='ubnt')
    else:
        credentials = dict(username=base_url.username, password=base_url.password)
        base_url._replace(netloc=base_url.netloc.replace(f'{base_url.username}:{base_url.password}@', ''))

    base_url = base_url.geturl()
    return (base_url, credentials)


def power_set(host, port, index, value):
    index = int(index)
    value = 1 if value else 0

    (base_url, credentials) = get_credentials(host)

    s = requests.Session()

    login(s, base_url, credentials)
    r = s.put(urljoin(base_url, f'/sensors/{index}/'), data=dict(output=value))
    if r.status_code == 200 and r.headers['Content-Type'] == 'application/json':
        j = r.json()
        if j['status'] != 'success':
            raise ExecutionError(f"unexpected API status code: '{j['status']}', response JSON: {j}")
    else:
        raise ExecutionError(f"unexpected http response: code {r.status_code}, content type '{r.headers['Content-Type']}' and content: '{r.text}'")


def power_get(host, port, index):
    index = int(index)

    s = requests.Session()

    (base_url, credentials) = get_credentials(host)

    login(s, base_url, credentials)

    r = s.get(urljoin(base_url, '/mfi/sensors.cgi'))
    if r.status_code == 200 and r.headers['Content-Type'] == 'application/json':
        j = r.json()
        if j['status'] != 'success':
            raise ExecutionError(f"unexpected API status code: '{j['status']}', response JSON: {j}")

        port = next(filter(lambda s: s['port'] == index, j['sensors']), None)
        if port is None:
            raise ExecutionError(f"port index '{index}' not found, available indices: '{[s['port'] for s in j['sensors']]}'")

        if port['output'] == 0:
            return False
        elif port['output'] == 1:
            return True
        else:
            raise ExecutionError(f"unexpected port output value: '{port['output']}'")
    else:
        raise ExecutionError(f"unexpected http response: code {r.status_code}, content type '{r.headers['Content-Type']}' and content: '{r.text}'")

    return False




#!/usr/bin/env python3
"""Capture Channel D1 of an fx2lafw device for 1 second."""

import sys
import time
import logging

from labgrid import Environment
from labgrid.logging import basicConfig

# enable info logging
basicConfig(level=logging.INFO)

env = Environment(sys.argv[1])
target = env.get_target("main")

sigrok = target.get_driver("SigrokDriver")
sigrok.capture("test.cap")
time.sleep(1)
sigrok.stop()



import enum

import attr

from ..factory import target_factory
from .common import Strategy, StrategyError, never_retry
from ..driver.dockerdriver import DockerDriver
from ..step import step


class Status(enum.Enum):
    """The possible states of a docker container"""

    unknown = 0
    gone = 1
    accessible = 2


@target_factory.reg_driver
@attr.s(eq=False)
class DockerStrategy(Strategy):
    """
    DockerStrategy enables the user to directly transition to a state
    where a fresh docker container has been created and is
    ready for access (e.g. shell access via SSH if the docker image
    runs an SSH daemon).
    """

    bindings = {
        "docker_driver": DockerDriver,
    }

    status = attr.ib(default=Status.unknown)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    @never_retry
    @step(args=["status"])
    def transition(self, status):
        if not isinstance(status, Status):
            status = Status[status]
        if status == self.status:
            return  # nothing to do
        elif status == Status.accessible:
            self.target.activate(self.docker_driver)
            self.docker_driver.on()
        elif status == Status.gone:
            self.target.activate(self.docker_driver)
            self.docker_driver.off()
            self.target.deactivate(self.docker_driver)
        else:
            raise StrategyError(f"no transition found from {self.status} to {status}")
        self.status = status



import abc


class ConsoleProtocol(abc.ABC):
    """Abstract class for the ConsoleProtocol"""

    @abc.abstractmethod
    def read(self):
        """
        Read data from underlying port
        """
        raise NotImplementedError

    @abc.abstractmethod
    def write(self, data: bytes):
        """
        Write data to underlying port
        """
        raise NotImplementedError

    def sendline(self, line: str):
        raise NotImplementedError

    def sendcontrol(self, char: str):
        raise NotImplementedError

    def expect(self, pattern: str):
        raise NotImplementedError

    class Client(abc.ABC):
        @abc.abstractmethod
        def get_console_matches(self):
            raise NotImplementedError

        @abc.abstractmethod
        def notify_console_match(self, pattern, match):
            raise NotImplementedError



import attr

from ..factory import target_factory
from .common import Resource


@target_factory.reg_resource
@attr.s(eq=False)
class OneWirePIO(Resource):
    """This resource describes a Onewire PIO Port.

    Args:
        host (str): hostname of the owserver e.g. localhost:4304
        path (str): path to the port on the owserver e.g. 29.7D6913000000/PIO.0
        invert (bool): optional, whether the logic level is be inverted (active-low)"""

    host = attr.ib(validator=attr.validators.instance_of(str))
    path = attr.ib(validator=attr.validators.instance_of(str))
    invert = attr.ib(default=False, validator=attr.validators.instance_of(bool))



import pytest

import os
from labgrid.util.agents.sysfsgpio import GpioDigitalOutput
from tempfile import TemporaryDirectory

class TestGpioAgent:

    class GpioDigitalOutputMock(GpioDigitalOutput):
        def __init__(self, **kwargs):
            index = kwargs['index']
            self.sysfs_mock_directory = TemporaryDirectory()
            GpioDigitalOutput._gpio_sysfs_path_prefix = self.sysfs_mock_directory.name
            GpioDigitalOutput._buffered_file_access = True
            export_file_path = os.path.join(self.sysfs_mock_directory.name, 'export')
            os.mknod(export_file_path)
            # Since there is no real device, writing to `export` does not create a corresponding
            # control directory. Expect a value error to be raised.
            with pytest.raises(ValueError, match='Device not found'):
                super().__init__(**kwargs)
            with open(export_file_path, mode='r') as export_file:
                export_content = export_file.readline()
                assert export_content == str(index)
            self.gpio_line_directory = os.path.join(self.sysfs_mock_directory.name, f'gpio{index}')
            os.mkdir(self.gpio_line_directory)
            for control_file in ['direction', 'value']:
                control_file_path = os.path.join(self.gpio_line_directory, control_file)
                print(f'creating control file `{control_file_path}`')
                os.mknod(control_file_path)
            super().__init__(**kwargs)

        def set(self, val):
            # new values are available at the beginning of the virtual file
            # emulate this behavior
            os.lseek(self.gpio_sysfs_value_fd, 0, os.SEEK_SET)
            super().set(val)

    def test_instantiation(self):
        gpio_line = TestGpioAgent.GpioDigitalOutputMock(index=13)
        assert isinstance(gpio_line, GpioDigitalOutput)

    def test_set(self):
        gpio_line = TestGpioAgent.GpioDigitalOutputMock(index=13)
        for val in [True, False, True, False]:
            gpio_line.set(val)
            assert gpio_line.get() == val



from pysnmp import hlapi
from ..driver.exception import ExecutionError


class SimpleSNMP:
    """A class that helps wrap pysnmp"""
    def __init__(self, host, community, port=161):
        if port is None:
            port = 161

        self.engine = hlapi.SnmpEngine()
        self.transport = hlapi.UdpTransportTarget((host, port))
        self.community = hlapi.CommunityData(community, mpModel=0)
        self.context = hlapi.ContextData()

    def get(self, oid):
        g = hlapi.getCmd(self.engine, self.community, self.transport,
            self.context, hlapi.ObjectType(hlapi.ObjectIdentity(oid)),
            lookupMib=False)

        error_indication, error_status, _, res = next(g)
        if error_indication or error_status:
            raise ExecutionError("Failed to get SNMP value.")
        return res[0][1]

    def set(self, oid, value):
        identify = hlapi.ObjectType(hlapi.ObjectIdentity(oid),
                   hlapi.Integer(value))
        g = hlapi.setCmd(self.engine, self.community, self.transport,
            self.context, identify, lookupMib=False)
        next(g)



import attr

from ..factory import target_factory
from .common import NetworkResource, Resource


@target_factory.reg_resource
@attr.s(eq=False)
class Flashrom(Resource):
    """Programmer is the programmer parameter described in man(8) of flashrom"""

    programmer = attr.ib(validator=attr.validators.instance_of(str))


@target_factory.reg_resource
@attr.s(eq=False)
class NetworkFlashrom(NetworkResource):
    programmer = attr.ib(validator=attr.validators.instance_of(str))
